"""Add models for OAuth and tweak AuditLog

Revision ID: 8d7729f5ed1f
Revises: 2b8bf1c3378e
Create Date: 2022-09-07 17:14:35.990594

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

import fief

# revision identifiers, used by Alembic.
revision = "8d7729f5ed1f"
down_revision = "2b8bf1c3378e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_oauth_providers",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("provider", sa.String(length=255), nullable=False),
        sa.Column("client_id", sa.Text(), nullable=False),
        sa.Column("client_secret", sa.Text(), nullable=False),
        sa.Column("scopes", sa.JSON(), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("openid_configuration_endpoint", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_oauth_providers_created_at"),
        "fief_oauth_providers",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_oauth_providers_updated_at"),
        "fief_oauth_providers",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_oauth_accounts",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("access_token", sa.Text(), nullable=False),
        sa.Column(
            "expires_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=True,
        ),
        sa.Column("refresh_token", sa.Text(), nullable=True),
        sa.Column("account_id", sa.String(length=512), nullable=False),
        sa.Column("account_email", sa.String(length=512), nullable=True),
        sa.Column("oauth_provider_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=True),
        sa.Column("tenant_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["oauth_provider_id"], ["fief_oauth_providers.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["tenant_id"], ["fief_tenants.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_id"], ["fief_users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("oauth_provider_id", "user_id"),
    )
    op.create_index(
        op.f("ix_fief_oauth_accounts_account_id"),
        "fief_oauth_accounts",
        ["account_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_oauth_accounts_created_at"),
        "fief_oauth_accounts",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_oauth_accounts_updated_at"),
        "fief_oauth_accounts",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_oauth_sessions",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("token", sa.String(length=255), nullable=False),
        sa.Column("redirect_uri", sa.Text(), nullable=False),
        sa.Column("oauth_provider_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("login_session_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("oauth_account_id", fief.models.generics.GUID(), nullable=True),
        sa.Column("tenant_id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "expires_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["login_session_id"], ["fief_login_sessions.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["oauth_account_id"], ["fief_oauth_accounts.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["oauth_provider_id"], ["fief_oauth_providers.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["tenant_id"], ["fief_tenants.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_oauth_sessions_created_at"),
        "fief_oauth_sessions",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_oauth_sessions_expires_at"),
        "fief_oauth_sessions",
        ["expires_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_oauth_sessions_token"),
        "fief_oauth_sessions",
        ["token"],
        unique=True,
    )
    op.create_index(
        op.f("ix_fief_oauth_sessions_updated_at"),
        "fief_oauth_sessions",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_registration_sessions",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("token", sa.String(length=255), nullable=False),
        sa.Column("flow", sa.String(length=255), nullable=False),
        sa.Column("email", sa.String(length=320), nullable=True),
        sa.Column("oauth_account_id", fief.models.generics.GUID(), nullable=True),
        sa.Column("tenant_id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "expires_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["oauth_account_id"], ["fief_oauth_accounts.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["tenant_id"], ["fief_tenants.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_registration_sessions_created_at"),
        "fief_registration_sessions",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_registration_sessions_expires_at"),
        "fief_registration_sessions",
        ["expires_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_registration_sessions_token"),
        "fief_registration_sessions",
        ["token"],
        unique=True,
    )
    op.create_index(
        op.f("ix_fief_registration_sessions_updated_at"),
        "fief_registration_sessions",
        ["updated_at"],
        unique=False,
    )
    op.add_column(
        "fief_audit_logs",
        sa.Column("object_id", fief.models.generics.GUID(), nullable=True),
    )
    op.add_column(
        "fief_audit_logs",
        sa.Column("object_class", sa.String(length=255), nullable=True),
    )
    op.add_column(
        "fief_audit_logs",
        sa.Column("admin_user_id", fief.models.generics.GUID(), nullable=True),
    )
    op.add_column(
        "fief_audit_logs",
        sa.Column("admin_api_key_id", fief.models.generics.GUID(), nullable=True),
    )
    op.create_index(
        op.f("ix_fief_audit_logs_admin_api_key_id"),
        "fief_audit_logs",
        ["admin_api_key_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_audit_logs_admin_user_id"),
        "fief_audit_logs",
        ["admin_user_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_audit_logs_object_class"),
        "fief_audit_logs",
        ["object_class"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_audit_logs_object_id"),
        "fief_audit_logs",
        ["object_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_audit_logs_subject_user_id"),
        "fief_audit_logs",
        ["subject_user_id"],
        unique=False,
    )

    connection = op.get_bind()
    if connection.dialect.name == "postgresql":
        op.drop_constraint(
            "fief_audit_logs_subject_user_id_fkey",
            "fief_audit_logs",
            type_="foreignkey",
        )
        op.drop_constraint(
            "fief_audit_logs_author_user_id_fkey", "fief_audit_logs", type_="foreignkey"
        )
        op.drop_column("fief_audit_logs", "author_user_id")
    elif connection.dialect.name == "mysql":
        op.drop_constraint(
            "fief_audit_logs_ibfk_1", "fief_audit_logs", type_="foreignkey"
        )
        op.drop_constraint(
            "fief_audit_logs_ibfk_2", "fief_audit_logs", type_="foreignkey"
        )
        op.drop_column("fief_audit_logs", "author_user_id")
    elif connection.dialect.name == "sqlite":
        with op.batch_alter_table("fief_audit_logs") as batch_op:
            batch_op.drop_column("author_user_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "fief_audit_logs",
        sa.Column(
            "author_user_id", postgresql.UUID(), autoincrement=False, nullable=True
        ),
    )
    op.create_foreign_key(
        "fief_audit_logs_subject_user_id_fkey",
        "fief_audit_logs",
        "fief_users",
        ["subject_user_id"],
        ["id"],
        ondelete="SET NULL",
    )
    op.create_foreign_key(
        "fief_audit_logs_author_user_id_fkey",
        "fief_audit_logs",
        "fief_users",
        ["author_user_id"],
        ["id"],
        ondelete="SET NULL",
    )
    op.drop_index(
        op.f("ix_fief_audit_logs_subject_user_id"), table_name="fief_audit_logs"
    )
    op.drop_index(op.f("ix_fief_audit_logs_object_id"), table_name="fief_audit_logs")
    op.drop_index(op.f("ix_fief_audit_logs_object_class"), table_name="fief_audit_logs")
    op.drop_index(
        op.f("ix_fief_audit_logs_admin_user_id"), table_name="fief_audit_logs"
    )
    op.drop_index(
        op.f("ix_fief_audit_logs_admin_api_key_id"), table_name="fief_audit_logs"
    )
    op.drop_column("fief_audit_logs", "admin_api_key_id")
    op.drop_column("fief_audit_logs", "admin_user_id")
    op.drop_column("fief_audit_logs", "object_class")
    op.drop_column("fief_audit_logs", "object_id")
    op.drop_index(
        op.f("ix_fief_registration_sessions_updated_at"),
        table_name="fief_registration_sessions",
    )
    op.drop_index(
        op.f("ix_fief_registration_sessions_token"),
        table_name="fief_registration_sessions",
    )
    op.drop_index(
        op.f("ix_fief_registration_sessions_expires_at"),
        table_name="fief_registration_sessions",
    )
    op.drop_index(
        op.f("ix_fief_registration_sessions_created_at"),
        table_name="fief_registration_sessions",
    )
    op.drop_table("fief_registration_sessions")
    op.drop_index(
        op.f("ix_fief_oauth_sessions_updated_at"), table_name="fief_oauth_sessions"
    )
    op.drop_index(
        op.f("ix_fief_oauth_sessions_token"), table_name="fief_oauth_sessions"
    )
    op.drop_index(
        op.f("ix_fief_oauth_sessions_expires_at"), table_name="fief_oauth_sessions"
    )
    op.drop_index(
        op.f("ix_fief_oauth_sessions_created_at"), table_name="fief_oauth_sessions"
    )
    op.drop_table("fief_oauth_sessions")
    op.drop_index(
        op.f("ix_fief_oauth_accounts_updated_at"), table_name="fief_oauth_accounts"
    )
    op.drop_index(
        op.f("ix_fief_oauth_accounts_created_at"), table_name="fief_oauth_accounts"
    )
    op.drop_index(
        op.f("ix_fief_oauth_accounts_account_id"), table_name="fief_oauth_accounts"
    )
    op.drop_table("fief_oauth_accounts")
    op.drop_index(
        op.f("ix_fief_oauth_providers_updated_at"), table_name="fief_oauth_providers"
    )
    op.drop_index(
        op.f("ix_fief_oauth_providers_created_at"), table_name="fief_oauth_providers"
    )
    op.drop_table("fief_oauth_providers")
    # ### end Alembic commands ###



from datetime import datetime, timezone

from sqlalchemy import select

from fief.models import SessionToken
from fief.repositories.base import BaseRepository, ExpiresAtMixin, UUIDRepositoryMixin


class SessionTokenRepository(
    BaseRepository[SessionToken],
    UUIDRepositoryMixin[SessionToken],
    ExpiresAtMixin[SessionToken],
):
    model = SessionToken

    async def get_by_token(
        self, token: str, *, fresh: bool = True
    ) -> SessionToken | None:
        statement = select(SessionToken).where(SessionToken.token == token)
        if fresh:
            statement = statement.where(
                SessionToken.expires_at > datetime.now(timezone.utc)
            )

        return await self.get_one_or_none(statement)



from fastapi import Depends, FastAPI
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.staticfiles import StaticFiles
from starlette.middleware.sessions import SessionMiddleware

from fief.apps.dashboard.dependencies import BaseContext, get_base_context
from fief.apps.dashboard.exception_handlers import exception_handlers
from fief.apps.dashboard.routers.api_keys import router as api_keys_router
from fief.apps.dashboard.routers.auth import router as auth_router
from fief.apps.dashboard.routers.clients import router as clients_router
from fief.apps.dashboard.routers.email_templates import router as email_templates_router
from fief.apps.dashboard.routers.oauth_providers import router as oauth_providers_router
from fief.apps.dashboard.routers.permissions import router as permissions_router
from fief.apps.dashboard.routers.roles import router as roles_router
from fief.apps.dashboard.routers.tenants import router as tenants_router
from fief.apps.dashboard.routers.themes import router as themes_router
from fief.apps.dashboard.routers.user_fields import router as user_fields_router
from fief.apps.dashboard.routers.users import router as users_router
from fief.apps.dashboard.routers.webhooks import router as webhooks_router
from fief.apps.dashboard.routers.workspaces import router as workspaces_router
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.middlewares.csrf import CSRFCookieSetterMiddleware
from fief.middlewares.security_headers import SecurityHeadersMiddleware
from fief.paths import STATIC_DIRECTORY
from fief.settings import settings
from fief.templates import templates

app = FastAPI(title="Fief Administration Dashboard", openapi_url=None)

app.add_middleware(SecurityHeadersMiddleware)
app.add_middleware(CSRFCookieSetterMiddleware)
app.add_middleware(GZipMiddleware)
app.add_middleware(
    SessionMiddleware,
    secret_key=settings.secret.get_secret_value(),
    session_cookie=settings.session_data_cookie_name,
    max_age=settings.session_data_cookie_lifetime_seconds,
    https_only=settings.session_data_cookie_secure,
)

app.include_router(permissions_router, prefix="/access-control/permissions")
app.include_router(roles_router, prefix="/access-control/roles")
app.include_router(api_keys_router, prefix="/api-keys")
app.include_router(auth_router, prefix="/auth")
app.include_router(clients_router, prefix="/clients")
app.include_router(email_templates_router, prefix="/customization/email-templates")
app.include_router(themes_router, prefix="/customization/themes")
app.include_router(oauth_providers_router, prefix="/oauth-providers")
app.include_router(tenants_router, prefix="/tenants")
app.include_router(user_fields_router, prefix="/user-fields")
app.include_router(users_router, prefix="/users")
app.include_router(webhooks_router, prefix="/webhooks")
app.include_router(workspaces_router, prefix="/workspaces")
app.mount("/static", StaticFiles(directory=STATIC_DIRECTORY), name="dashboard:static")

for exc, handler in exception_handlers.items():
    app.add_exception_handler(exc, handler)


@app.get(
    "/", name="dashboard:index", dependencies=[Depends(is_authenticated_admin_session)]
)
async def index(context: BaseContext = Depends(get_base_context)):
    return templates.TemplateResponse("admin/index.html", {**context})


__all__ = ["app"]



from fastapi import APIRouter, Depends, HTTPException, Response, status

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.tenant import get_paginated_tenants, get_tenant_by_id_or_404
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.errors import APIErrorCode
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Client, OAuthProvider, Tenant
from fief.repositories import (
    ClientRepository,
    OAuthProviderRepository,
    TenantRepository,
    ThemeRepository,
)
from fief.schemas.generics import PaginatedResults
from fief.services.webhooks.models import (
    ClientCreated,
    TenantCreated,
    TenantDeleted,
    TenantUpdated,
)

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/", name="tenants:list", response_model=PaginatedResults[schemas.tenant.Tenant]
)
async def list_tenants(
    paginated_tenants: PaginatedObjects[Tenant] = Depends(get_paginated_tenants),
) -> PaginatedResults[schemas.tenant.Tenant]:
    tenants, count = paginated_tenants
    return PaginatedResults(
        count=count,
        results=[schemas.tenant.Tenant.from_orm(tenant) for tenant in tenants],
    )


@router.get("/{id:uuid}", name="tenants:get", response_model=schemas.tenant.Tenant)
async def get_tenant(tenant: Tenant = Depends(get_tenant_by_id_or_404)) -> Tenant:
    return tenant


@router.post(
    "/",
    name="tenants:create",
    status_code=status.HTTP_201_CREATED,
    response_model=schemas.tenant.Tenant,
)
async def create_tenant(
    tenant_create: schemas.tenant.TenantCreate,
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
    client_repository: ClientRepository = Depends(
        get_workspace_repository(ClientRepository)
    ),
    theme_repository: ThemeRepository = Depends(
        get_workspace_repository(ThemeRepository)
    ),
    oauth_provider_repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.tenant.Tenant:
    if tenant_create.theme_id is not None:
        theme = await theme_repository.get_by_id(tenant_create.theme_id)
        if theme is None:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                detail=APIErrorCode.TENANT_CREATE_NOT_EXISTING_THEME,
            )

    oauth_providers: list[OAuthProvider] = []
    if tenant_create.oauth_providers is not None:
        for oauth_provider_id in tenant_create.oauth_providers:
            oauth_provider = await oauth_provider_repository.get_by_id(
                oauth_provider_id
            )
            if oauth_provider is None:
                raise HTTPException(
                    status.HTTP_400_BAD_REQUEST,
                    detail=APIErrorCode.TENANT_CREATE_NOT_EXISTING_OAUTH_PROVIDER,
                )
            oauth_providers.append(oauth_provider)

    slug = await repository.get_available_slug(tenant_create.name)
    tenant = Tenant(
        **tenant_create.dict(exclude={"oauth_providers"}),
        slug=slug,
        oauth_providers=oauth_providers,
    )

    tenant = await repository.create(tenant)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, tenant)
    trigger_webhooks(TenantCreated, tenant, schemas.tenant.Tenant)

    client = Client(
        name=f"{tenant.name}'s client",
        first_party=True,
        tenant=tenant,
        redirect_uris=["http://localhost:8000/docs/oauth2-redirect"],
    )
    await client_repository.create(client)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, client)
    trigger_webhooks(ClientCreated, client, schemas.client.Client)

    return schemas.tenant.Tenant.from_orm(tenant)


@router.patch("/{id:uuid}", name="tenants:update", response_model=schemas.tenant.Tenant)
async def update_tenant(
    tenant_update: schemas.tenant.TenantUpdate,
    tenant: Tenant = Depends(get_tenant_by_id_or_404),
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
    theme_repository: ThemeRepository = Depends(
        get_workspace_repository(ThemeRepository)
    ),
    oauth_provider_repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.tenant.Tenant:
    if tenant_update.theme_id is not None:
        theme = await theme_repository.get_by_id(tenant_update.theme_id)
        if theme is None:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                detail=APIErrorCode.TENANT_UPDATE_NOT_EXISTING_THEME,
            )

    if tenant_update.oauth_providers is not None:
        tenant.oauth_providers = []
        for oauth_provider_id in tenant_update.oauth_providers:
            oauth_provider = await oauth_provider_repository.get_by_id(
                oauth_provider_id
            )
            if oauth_provider is None:
                raise HTTPException(
                    status.HTTP_400_BAD_REQUEST,
                    detail=APIErrorCode.TENANT_UPDATE_NOT_EXISTING_OAUTH_PROVIDER,
                )
            tenant.oauth_providers.append(oauth_provider)

    tenant_update_dict = tenant_update.dict(
        exclude={"oauth_providers"}, exclude_unset=True
    )
    for field, value in tenant_update_dict.items():
        setattr(tenant, field, value)

    await repository.update(tenant)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, tenant)
    trigger_webhooks(TenantUpdated, tenant, schemas.tenant.Tenant)

    return schemas.tenant.Tenant.from_orm(tenant)


@router.delete(
    "/{id:uuid}",
    name="tenants:delete",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_tenant(
    tenant: Tenant = Depends(get_tenant_by_id_or_404),
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    await repository.delete(tenant)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, tenant)
    trigger_webhooks(TenantDeleted, tenant, schemas.tenant.Tenant)



from sqlalchemy import select

from fief.models import AdminAPIKey
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class AdminAPIKeyRepository(
    BaseRepository[AdminAPIKey], UUIDRepositoryMixin[AdminAPIKey]
):
    model = AdminAPIKey

    async def get_by_token(self, token: str) -> AdminAPIKey | None:
        statement = select(AdminAPIKey).where(AdminAPIKey.token == token)
        return await self.get_one_or_none(statement)



from fastapi import Cookie, Depends, Query
from pydantic import UUID4

from fief.dependencies.oauth_provider import get_oauth_providers
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.exceptions import OAuthException
from fief.locale import gettext_lazy as _
from fief.models import LoginSession, OAuthProvider, Tenant
from fief.repositories import LoginSessionRepository, TenantRepository
from fief.schemas.oauth import OAuthError
from fief.settings import settings


async def get_tenant_by_query(
    tenant_id: UUID4 = Query(..., alias="tenant"),
    tenant_repository: TenantRepository = Depends(
        get_workspace_repository(TenantRepository)
    ),
) -> Tenant:
    tenant = await tenant_repository.get_by_id(tenant_id)

    if tenant is None:
        raise OAuthException(
            OAuthError.get_invalid_tenant(_("Unknown tenant")), fatal=True
        )

    return tenant


async def get_optional_login_session_with_tenant_query(
    token: str | None = Cookie(None, alias=settings.login_session_cookie_name),
    login_session_repository: LoginSessionRepository = Depends(
        get_workspace_repository(LoginSessionRepository)
    ),
    tenant: Tenant = Depends(get_tenant_by_query),
) -> LoginSession | None:
    if token is None:
        return None

    login_session = await login_session_repository.get_by_token(token)
    if login_session is None:
        return None

    if login_session.client.tenant_id != tenant.id:
        return None

    return login_session


async def get_login_session_with_tenant_query(
    login_session: LoginSession
    | None = Depends(get_optional_login_session_with_tenant_query),
    tenant: Tenant = Depends(get_tenant_by_query),
    oauth_providers: list[OAuthProvider] | None = Depends(get_oauth_providers),
) -> LoginSession:
    if login_session is None:
        raise OAuthException(
            OAuthError.get_invalid_session(_("Invalid login session.")),
            oauth_providers=oauth_providers,
            tenant=tenant,
        )

    return login_session


async def get_oauth_provider(
    provider: UUID4 = Query(...),
    tenant: Tenant = Depends(get_tenant_by_query),
    oauth_providers: list[OAuthProvider] | None = Depends(get_oauth_providers),
) -> OAuthProvider:
    oauth_provider = tenant.get_oauth_provider(provider)
    if oauth_provider is None:
        raise OAuthException(
            OAuthError.get_invalid_provider(_("Unknown OAuth provider")),
            oauth_providers=oauth_providers,
            tenant=tenant,
        )

    return oauth_provider



from fastapi import APIRouter, Depends, Request, status

from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.webhook import WebhookCreateForm, WebhookUpdateForm
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.webhook import (
    get_paginated_webhook_logs,
    get_paginated_webhooks,
    get_webhook_by_id_or_404,
    get_webhook_log_by_id_and_webhook_or_404,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Webhook, WebhookLog
from fief.repositories import WebhookRepository
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("URL", "url", "url_column", ordering="url"),
        DatatableColumn(
            "Created at", "created_at", "created_at_column", ordering="created_at"
        ),
        DatatableColumn("Events", "events", "events_column"),
        DatatableColumn("Actions", "actions", "actions_column"),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["url", "created_at", "events", "actions"])
    ),
    paginated_webhooks: PaginatedObjects[Webhook] = Depends(get_paginated_webhooks),
):
    webhooks, count = paginated_webhooks
    return {
        "webhooks": webhooks,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


@router.get("/", name="dashboard.webhooks:list")
async def list_webhooks(
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/webhooks/list.html", {**context, **list_context}
    )


@router.get("/{id:uuid}", name="dashboard.webhooks:get")
async def get_webhook(
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/webhooks/get.html",
        {**context, **list_context, "webhook": webhook},
    )


@router.api_route("/create", methods=["GET", "POST"], name="dashboard.webhooks:create")
async def create_webhook(
    request: Request,
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    form_helper = FormHelper(
        WebhookCreateForm,
        "admin/webhooks/create.html",
        request=request,
        context={**context, **list_context},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        webhook = Webhook()
        form.populate_obj(webhook)

        webhook = await repository.create(webhook)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, webhook)

        return templates.TemplateResponse(
            "admin/webhooks/secret.html",
            {**context, **list_context, "webhook": webhook, "secret": webhook.secret},
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(webhook.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/edit", methods=["GET", "POST"], name="dashboard.webhooks:update"
)
async def update_webhook(
    request: Request,
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    form_helper = FormHelper(
        WebhookUpdateForm,
        "admin/webhooks/edit.html",
        object=webhook,
        request=request,
        context={**context, **list_context, "webhook": webhook},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        form.populate_obj(webhook)

        await repository.update(webhook)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, webhook)

        return HXRedirectResponse(
            request.url_for("dashboard.webhooks:get", id=webhook.id)
        )

    return await form_helper.get_response()


@router.post("/{id:uuid}/secret", name="dashboard.webhooks:secret")
async def regenerate_webhook_secret(
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    secret = webhook.regenerate_secret()
    await repository.update(webhook)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, webhook)

    return templates.TemplateResponse(
        "admin/webhooks/secret.html",
        {**context, **list_context, "webhook": webhook, "secret": secret},
        headers={"X-Fief-Object-Id": str(webhook.id)},
    )


@router.api_route(
    "/{id:uuid}/delete", methods=["GET", "DELETE"], name="dashboard.webhooks:delete"
)
async def delete_webhook(
    request: Request,
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    if request.method == "DELETE":
        await repository.delete(webhook)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, webhook)

        return HXRedirectResponse(
            request.url_for("dashboard.webhooks:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )
    else:
        return templates.TemplateResponse(
            "admin/webhooks/delete.html",
            {**context, **list_context, "webhook": webhook},
        )


async def get_logs_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn(
            "Delivered at", "delivered_at", "created_at_column", ordering="created_at"
        ),
        DatatableColumn("Event", "event", "event_column", ordering="event"),
        DatatableColumn("Success", "success", "success_column", ordering="success"),
        DatatableColumn("Attempt", "attempt", "attempt_column", ordering="attempt"),
        DatatableColumn("Error", "error", "error_column", ordering="error_type"),
        DatatableColumn("Actions", "actions", "actions_column"),
    ]


async def get_logs_list_context(
    columns: list[DatatableColumn] = Depends(get_logs_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(
            ["delivered_at", "event", "success", "attempt", "error", "actions"]
        )
    ),
    paginated_webhook_logs: PaginatedObjects[WebhookLog] = Depends(
        get_paginated_webhook_logs
    ),
):
    webhook_logs, count = paginated_webhook_logs
    return {
        "webhook_logs": webhook_logs,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


@router.get("/{id:uuid}/logs", name="dashboard.webhooks:logs")
async def list_webhook_logs(
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    list_context=Depends(get_logs_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/webhooks/logs/list.html",
        {**context, **list_context, "webhook": webhook},
    )


@router.get("/{id:uuid}/logs/{log_id:uuid}", name="dashboard.webhooks:log")
async def get_webhook_log(
    webhook_log: WebhookLog = Depends(get_webhook_log_by_id_and_webhook_or_404),
    list_context=Depends(get_logs_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/webhooks/logs/get.html",
        {**context, **list_context, "webhook_log": webhook_log},
    )



from pydantic import UUID4
from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import Select

from fief.models import UserRole
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class UserRoleRepository(BaseRepository[UserRole], UUIDRepositoryMixin[UserRole]):
    model = UserRole

    def get_by_user_statement(self, user: UUID4) -> Select:
        statement = (
            select(UserRole)
            .where(UserRole.user_id == user)
            .options(joinedload(UserRole.role))
        )

        return statement

    async def get_by_role_and_user(self, user: UUID4, role: UUID4) -> UserRole | None:
        return await self.get_one_or_none(
            select(UserRole)
            .where(UserRole.user_id == user, UserRole.role_id == role)
            .options(joinedload(UserRole.role))
        )

    async def get_by_role(self, role: UUID4) -> list[UserRole]:
        return await self.list(select(UserRole).where(UserRole.role_id == role))



from fastapi import Depends

from fief.dependencies.admin_api_key import get_optional_admin_api_key
from fief.dependencies.admin_session import get_optional_admin_session_token
from fief.dependencies.current_workspace import get_current_workspace
from fief.logger import AuditLogger, logger
from fief.models import AdminAPIKey, AdminSessionToken, Workspace


async def get_audit_logger(
    workspace: Workspace = Depends(get_current_workspace),
    admin_session_token: AdminSessionToken
    | None = Depends(get_optional_admin_session_token),
    admin_api_key: AdminAPIKey | None = Depends(get_optional_admin_api_key),
) -> AuditLogger:
    return AuditLogger(
        logger,  # type: ignore
        workspace.id,
        admin_user_id=admin_session_token.user_id if admin_session_token else None,
        admin_api_key_id=admin_api_key.id if admin_api_key else None,
    )



from pydantic import UUID4
from sqlalchemy import select

from fief.models import User
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class UserRepository(BaseRepository[User], UUIDRepositoryMixin[User]):
    model = User

    async def get_one_by_tenant(self, tenant: UUID4) -> User | None:
        statement = (
            select(User)
            .where(User.tenant_id == tenant)
            .order_by(User.created_at)
            .limit(1)
        )
        return await self.get_one_or_none(statement)

    async def count_all(self) -> int:
        statement = select(User)
        return await self._count(statement)

    async def count_by_tenant(self, tenant: UUID4) -> int:
        statement = select(User).where(User.tenant_id == tenant)
        return await self._count(statement)



import secrets

from pydantic import UUID4
from sqlalchemy import ForeignKey, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.sqltypes import String

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, ExpiresAt, UUIDModel
from fief.models.login_session import LoginSession
from fief.models.oauth_account import OAuthAccount
from fief.models.oauth_provider import OAuthProvider
from fief.models.tenant import Tenant
from fief.settings import settings


class OAuthSession(UUIDModel, CreatedUpdatedAt, ExpiresAt, WorkspaceBase):
    __tablename__ = "oauth_sessions"
    __lifetime_seconds__ = settings.login_session_lifetime_seconds

    token: Mapped[str] = mapped_column(
        String(length=255),
        default=secrets.token_urlsafe,
        nullable=False,
        index=True,
        unique=True,
    )
    redirect_uri: Mapped[str] = mapped_column(Text, nullable=False)

    oauth_provider_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(OAuthProvider.id, ondelete="CASCADE"), nullable=False
    )
    oauth_provider: Mapped[OAuthProvider] = relationship("OAuthProvider", lazy="joined")

    login_session_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(LoginSession.id, ondelete="CASCADE"), nullable=False
    )
    login_session: Mapped[LoginSession] = relationship("LoginSession")

    oauth_account_id: Mapped[UUID4 | None] = mapped_column(
        GUID, ForeignKey(OAuthAccount.id, ondelete="CASCADE"), nullable=True
    )
    oauth_account: Mapped[OAuthAccount | None] = relationship(
        "OAuthAccount", lazy="joined"
    )

    tenant_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Tenant.id, ondelete="CASCADE"), nullable=False
    )
    tenant: Mapped[Tenant] = relationship("Tenant", lazy="joined")



from fief.tasks.base import SendTask, send_task
from fief.tasks.cleanup import cleanup, cleanup_workspace
from fief.tasks.count_users import count_users, count_users_workspace
from fief.tasks.forgot_password import on_after_forgot_password
from fief.tasks.register import on_after_register
from fief.tasks.roles import on_role_updated
from fief.tasks.user_permissions import on_user_role_created, on_user_role_deleted
from fief.tasks.webhooks import deliver_webhook, trigger_webhooks

__all__ = [
    "send_task",
    "SendTask",
    "cleanup",
    "cleanup_workspace",
    "count_users",
    "count_users_workspace",
    "on_after_forgot_password",
    "on_after_register",
    "on_role_updated",
    "on_user_role_created",
    "on_user_role_deleted",
    "deliver_webhook",
    "trigger_webhooks",
]



from fastapi import APIRouter, Depends, Header, Request, status

from fief import schemas
from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.role import RoleCreateForm, RoleUpdateForm
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.role import get_paginated_roles, get_role_by_id_or_404
from fief.dependencies.tasks import get_send_task
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Role, Workspace
from fief.repositories import PermissionRepository, RoleRepository
from fief.services.webhooks.models import RoleCreated, RoleDeleted, RoleUpdated
from fief.tasks import SendTask, on_role_updated
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Name", "name", "name_column", ordering="name"),
        DatatableColumn(
            "Granted by default",
            "granted_by_default",
            "granted_by_default_column",
            ordering="granted_by_default",
        ),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["name", "granted_by_default"])
    ),
    paginated_roles: PaginatedObjects[Role] = Depends(get_paginated_roles),
):
    roles, count = paginated_roles
    return {
        "roles": roles,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


async def get_list_template(hx_combobox: bool = Header(False)) -> str:
    if hx_combobox:
        return "admin/roles/list_combobox.html"
    return "admin/roles/list.html"


@router.get("/", name="dashboard.roles:list")
async def list_roles(
    template: str = Depends(get_list_template),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(template, {**context, **list_context})


@router.get("/{id:uuid}", name="dashboard.roles:get")
async def get_role(
    role: Role = Depends(get_role_by_id_or_404),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/roles/get.html",
        {**context, **list_context, "role": role},
    )


@router.api_route("/create", methods=["GET", "POST"], name="dashboard.roles:create")
async def create_role(
    request: Request,
    repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    permission_repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        RoleCreateForm,
        "admin/roles/create.html",
        request=request,
        context={**context, **list_context},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        role = Role()
        form.populate_obj(role)
        role.permissions = []
        for permission_id in form.data["permissions"]:
            permission = await permission_repository.get_by_id(permission_id)
            if permission is None:
                form.permissions.errors.append("Unknown permission.")
                return await form_helper.get_error_response(
                    "Unknown permission.", "unknown_permission"
                )
            role.permissions.append(permission)

        role = await repository.create(role)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, role)
        trigger_webhooks(RoleCreated, role, schemas.role.Role)

        return HXRedirectResponse(
            request.url_for("dashboard.roles:get", id=role.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(role.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/edit",
    methods=["GET", "POST"],
    name="dashboard.roles:update",
)
async def update_role(
    request: Request,
    role: Role = Depends(get_role_by_id_or_404),
    repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    permission_repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    send_task: SendTask = Depends(get_send_task),
    workspace: Workspace = Depends(get_current_workspace),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        RoleUpdateForm,
        "admin/roles/edit.html",
        object=role,
        request=request,
        context={**context, **list_context, "role": role},
    )
    form = await form_helper.get_form()
    form.permissions.choices = [
        (permission.id, permission.codename) for permission in role.permissions
    ]

    if await form_helper.is_submitted_and_valid():
        old_permissions = {permission.id for permission in role.permissions}

        role.permissions = []
        for permission_id in form.data["permissions"]:
            permission = await permission_repository.get_by_id(permission_id)
            if permission is None:
                form.permissions.errors.append("Unknown permission.")
                return await form_helper.get_error_response(
                    "Unknown permission.", "unknown_permission"
                )
            role.permissions.append(permission)
        new_permissions = {permission.id for permission in role.permissions}

        del form.permissions
        form.populate_obj(role)

        await repository.update(role)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, role)
        trigger_webhooks(RoleUpdated, role, schemas.role.Role)

        added_permissions = new_permissions - old_permissions
        deleted_permissions = old_permissions - new_permissions
        send_task(
            on_role_updated,
            str(role.id),
            set(map(str, added_permissions)),
            set(map(str, deleted_permissions)),
            str(workspace.id),
        )

        return HXRedirectResponse(request.url_for("dashboard.roles:get", id=role.id))

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/delete",
    methods=["GET", "DELETE"],
    name="dashboard.roles:delete",
)
async def delete_role(
    request: Request,
    role: Role = Depends(get_role_by_id_or_404),
    repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    if request.method == "DELETE":
        await repository.delete(role)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, role)
        trigger_webhooks(RoleDeleted, role, schemas.role.Role)

        return HXRedirectResponse(
            request.url_for("dashboard.roles:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )
    else:
        return templates.TemplateResponse(
            "admin/roles/delete.html",
            {**context, **list_context, "role": role},
        )



from fastapi import APIRouter, Depends, Query, Request

from fief import schemas
from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.email_template import EmailTemplateUpdateForm
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.db import AsyncSession
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.current_workspace import get_current_workspace_session
from fief.dependencies.email_template import (
    get_email_template_by_id_or_404,
    get_paginated_email_templates,
)
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, EmailTemplate
from fief.repositories import EmailTemplateRepository
from fief.services.email_template.contexts import EMAIL_TEMPLATE_CONTEXT_CLASS_MAP
from fief.services.email_template.renderers import (
    EmailSubjectRenderer,
    EmailTemplateRenderer,
)
from fief.services.email_template.types import EmailTemplateType
from fief.services.webhooks.models import EmailTemplateUpdated
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Type", "type", "type_column", ordering="type"),
        DatatableColumn("Subject", "subject", "subject_column"),
        DatatableColumn("Actions", "actions", "actions_column"),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["type", "subject", "actions"])
    ),
    paginated_email_templates: PaginatedObjects[EmailTemplate] = Depends(
        get_paginated_email_templates
    ),
):
    email_templates, count = paginated_email_templates
    return {
        "email_templates": email_templates,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


@router.get("/", name="dashboard.email_templates:list")
async def list_email_templates(
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/email_templates/list.html", {**context, **list_context}
    )


@router.api_route(
    "/{id:uuid}/edit",
    methods=["GET", "POST"],
    name="dashboard.email_templates:update",
)
async def update_email_template(
    request: Request,
    preview: bool = Query(False),
    email_template: EmailTemplate = Depends(get_email_template_by_id_or_404),
    repository: EmailTemplateRepository = Depends(
        get_workspace_repository(EmailTemplateRepository)
    ),
    session: AsyncSession = Depends(get_current_workspace_session),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        EmailTemplateUpdateForm,
        "admin/email_templates/edit.html",
        object=email_template,
        request=request,
        context={**context, "email_template": email_template},
    )

    context_class = EMAIL_TEMPLATE_CONTEXT_CLASS_MAP[email_template.type]
    sample_context = await context_class.create_sample_context(session)

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        if not preview:
            form.populate_obj(email_template)
            await repository.update(email_template)
            audit_logger.log_object_write(
                AuditLogMessage.OBJECT_UPDATED, email_template
            )
            trigger_webhooks(
                EmailTemplateUpdated,
                email_template,
                schemas.email_template.EmailTemplate,
            )
            return HXRedirectResponse(request.url_for("dashboard.email_templates:list"))

        email_template = EmailTemplate(
            type=email_template.type,
            subject=form.data["subject"],
            content=form.data["content"],
        )
        email_subject_renderer = EmailSubjectRenderer(
            repository, templates_overrides={email_template.type: email_template}
        )
        email_template_renderer = EmailTemplateRenderer(
            repository, templates_overrides={email_template.type: email_template}
        )
        subject = await email_subject_renderer.render(EmailTemplateType[email_template.type], sample_context)  # type: ignore
        content = await email_template_renderer.render(EmailTemplateType[email_template.type], sample_context)  # type: ignore
    else:
        email_subject_renderer = EmailSubjectRenderer(repository)
        email_template_renderer = EmailTemplateRenderer(repository)
        subject = await email_subject_renderer.render(EmailTemplateType[email_template.type], sample_context)  # type: ignore
        content = await email_template_renderer.render(EmailTemplateType[email_template.type], sample_context)  # type: ignore

    form_helper.context["subject_preview"] = subject
    form_helper.context["content_preview"] = content

    return await form_helper.get_response()



from datetime import datetime, timezone
from typing import cast

from fastapi import APIRouter, Depends, HTTPException, Response, status
from fastapi.exceptions import RequestValidationError
from fastapi_users.exceptions import UserNotExists
from httpx_oauth.oauth2 import RefreshTokenError, RefreshTokenNotSupportedError
from pydantic import UUID4, ValidationError

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.oauth_provider import (
    get_oauth_provider_by_id_or_404,
    get_paginated_oauth_providers,
)
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.users import UserManager, get_user_manager
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.errors import APIErrorCode
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, OAuthProvider
from fief.models.oauth_account import OAuthAccount
from fief.repositories import OAuthAccountRepository, OAuthProviderRepository
from fief.schemas.generics import PaginatedResults
from fief.services.oauth_provider import get_oauth_provider_service
from fief.services.webhooks.models import (
    OAuthProviderCreated,
    OAuthProviderDeleted,
    OAuthProviderUpdated,
)

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/",
    name="oauth_providers:list",
    response_model=PaginatedResults[schemas.oauth_provider.OAuthProvider],
)
async def list_oauth_providers(
    paginated_oauth_providers: PaginatedObjects[OAuthProvider] = Depends(
        get_paginated_oauth_providers
    ),
) -> PaginatedResults[schemas.oauth_provider.OAuthProvider]:
    oauth_providers, count = paginated_oauth_providers
    return PaginatedResults(
        count=count,
        results=[
            schemas.oauth_provider.OAuthProvider.from_orm(oauth_provider)
            for oauth_provider in oauth_providers
        ],
    )


@router.get(
    "/{id:uuid}",
    name="oauth_providers:get",
    response_model=schemas.oauth_provider.OAuthProvider,
)
async def get_oauth_provider(
    oauth_provider: OAuthProvider = Depends(get_oauth_provider_by_id_or_404),
) -> OAuthProvider:
    return oauth_provider


@router.post(
    "/",
    name="oauth_providers:create",
    response_model=schemas.oauth_provider.OAuthProvider,
    status_code=status.HTTP_201_CREATED,
)
async def create_oauth_provider(
    oauth_provider_create: schemas.oauth_provider.OAuthProviderCreate,
    repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.oauth_provider.OAuthProvider:
    oauth_provider = OAuthProvider(**oauth_provider_create.dict())
    oauth_provider = await repository.create(oauth_provider)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, oauth_provider)
    trigger_webhooks(
        OAuthProviderCreated,
        oauth_provider,
        schemas.oauth_provider.OAuthProvider,
    )

    return schemas.oauth_provider.OAuthProvider.from_orm(oauth_provider)


@router.patch(
    "/{id:uuid}",
    name="oauth_providers:update",
    response_model=schemas.oauth_provider.OAuthProvider,
)
async def update_oauth_provider(
    oauth_provider_update: schemas.oauth_provider.OAuthProviderUpdate,
    oauth_provider: OAuthProvider = Depends(get_oauth_provider_by_id_or_404),
    repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.oauth_provider.OAuthProvider:
    oauth_provider_update_dict = oauth_provider_update.dict(exclude_unset=True)

    try:
        oauth_provider_update_provider = (
            schemas.oauth_provider.OAuthProviderUpdateProvider.from_orm(oauth_provider)
        )
        schemas.oauth_provider.OAuthProviderUpdateProvider(
            **oauth_provider_update_provider.copy(
                update=oauth_provider_update_dict
            ).dict()
        )
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e

    for field, value in oauth_provider_update_dict.items():
        setattr(oauth_provider, field, value)

    await repository.update(oauth_provider)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, oauth_provider)
    trigger_webhooks(
        OAuthProviderUpdated,
        oauth_provider,
        schemas.oauth_provider.OAuthProvider,
    )

    return schemas.oauth_provider.OAuthProvider.from_orm(oauth_provider)


@router.delete(
    "/{id:uuid}",
    name="oauth_providers:delete",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_oauth_provider(
    oauth_provider: OAuthProvider = Depends(get_oauth_provider_by_id_or_404),
    repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    await repository.delete(oauth_provider)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, oauth_provider)
    trigger_webhooks(
        OAuthProviderDeleted,
        oauth_provider,
        schemas.oauth_provider.OAuthProvider,
    )


@router.get(
    "/{id:uuid}/access-token/{user_id:uuid}",
    name="oauth_providers:get_user_access_token",
    response_model=schemas.oauth_account.OAuthAccountAccessToken,
)
async def get_user_access_token(
    user_id: UUID4,
    oauth_provider: OAuthProvider = Depends(get_oauth_provider_by_id_or_404),
    oauth_account_repository: OAuthAccountRepository = Depends(
        get_workspace_repository(OAuthAccountRepository)
    ),
    user_manager: UserManager = Depends(get_user_manager),
    audit_logger: AuditLogger = Depends(get_audit_logger),
) -> OAuthAccount:
    try:
        user = await user_manager.get(user_id)
    except UserNotExists as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND) from e

    oauth_account = await oauth_account_repository.get_by_provider_and_user(
        oauth_provider.id, user.id
    )
    if oauth_account is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    if oauth_account.is_expired():
        oauth_provider_service = get_oauth_provider_service(oauth_provider)
        try:
            access_token_dict = await oauth_provider_service.refresh_token(
                cast(str, oauth_account.refresh_token)
            )
        except RefreshTokenNotSupportedError as e:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=APIErrorCode.OAUTH_PROVIDER_REFRESH_TOKEN_NOT_SUPPORTED,
            ) from e
        except RefreshTokenError as e:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=APIErrorCode.OAUTH_PROVIDER_REFRESH_TOKEN_ERROR,
            ) from e
        oauth_account.access_token = access_token_dict["access_token"]  # type: ignore
        try:
            oauth_account.expires_at = datetime.fromtimestamp(
                access_token_dict["expires_at"], tz=timezone.utc
            )
        except KeyError:
            oauth_account.expires_at = None
        await oauth_account_repository.update(oauth_account)

    audit_logger(
        AuditLogMessage.OAUTH_PROVIDER_USER_ACCESS_TOKEN_GET,
        subject_user_id=user.id,
        oauth_provider_id=str(oauth_provider.id),
    )

    return oauth_account



from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger

from fief import tasks

scheduler = BlockingScheduler()


def schedule():
    scheduler = BlockingScheduler()
    scheduler.add_job(
        tasks.cleanup.send,
        CronTrigger.from_crontab("0 0 * * *"),
    )
    scheduler.add_job(
        tasks.count_users.send,
        CronTrigger.from_crontab("0 0 * * *"),
    )
    try:
        scheduler.start()
    except KeyboardInterrupt:
        scheduler.shutdown()



"""Add database_ssl_mode to Workspace

Revision ID: 4c2bb65882d7
Revises: 696e383aabf2
Create Date: 2022-06-09 09:57:28.046792

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "4c2bb65882d7"
down_revision = "696e383aabf2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "workspaces", sa.Column("database_ssl_mode", sa.Text(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workspaces", "database_ssl_mode")
    # ### end Alembic commands ###



import functools
import secrets
from collections.abc import Mapping
from typing import Any, Generic, TypeVar

import phonenumbers
import pycountry
import pytz
from fastapi import Request, status
from starlette.templating import _TemplateResponse
from wtforms import (
    BooleanField,
    DateField,
    DateTimeLocalField,
    Field,
    Form,
    FormField,
    HiddenField,
    IntegerField,
    PasswordField,
    SelectField,
    SelectMultipleField,
    StringField,
    TelField,
    validators,
    widgets,
)
from wtforms.csrf.core import CSRF
from wtforms.utils import unset_value

from fief.locale import get_translations
from fief.locale import gettext_lazy as _
from fief.middlewares.csrf import CSRF_ATTRIBUTE_NAME
from fief.models import UserField, UserFieldType
from fief.settings import settings
from fief.templates import templates


class CSRFCookieMissingRequest(TypeError):
    def __init__(self) -> None:
        super().__init__("Must provide a Request object for CSRF to work.")


class CSRFCookie(CSRF):
    def setup_form(self, form: Form):
        self.form_meta = form.meta
        return super().setup_form(form)

    def generate_csrf_token(self, csrf_token_field: Field):
        csrf_token: str | None = self.get_challenge_csrf_token()
        if csrf_token is None:
            csrf_token = secrets.token_urlsafe()
            # Will be catched by CSRFCookieSetterMiddleware and set in cookies
            self.request.scope[CSRF_ATTRIBUTE_NAME] = csrf_token

        return csrf_token

    def validate_csrf_token(self, form: Form, field: Field):
        challenge_csrf_token = self.get_challenge_csrf_token()
        if (
            field.data is None
            or challenge_csrf_token is None
            or not secrets.compare_digest(field.data, challenge_csrf_token)
        ):
            raise validators.ValidationError(field.gettext("CSRF failed."))

    def get_challenge_csrf_token(self) -> str | None:
        return self.request.cookies.get(self.form_meta.csrf_cookie_name)

    @property
    def request(self) -> Request:
        try:
            return getattr(self.form_meta, "request")
        except AttributeError as e:
            raise CSRFCookieMissingRequest() from e


class BaseForm(Form):
    class Meta:
        def get_translations(self, form):
            return get_translations(domain="wtforms")


class CSRFBaseForm(BaseForm):
    class Meta:
        csrf = settings.csrf_check_enabled
        csrf_class = CSRFCookie
        csrf_cookie_name = settings.csrf_cookie_name


F = TypeVar("F", bound=BaseForm)


class FormHelper(Generic[F]):
    def __init__(
        self,
        form_class: type[F],
        template: str,
        *,
        request: Request,
        object: Any | None = None,
        data: Any | None = None,
        context: dict | None = None,
    ):
        self.form_class = form_class
        self.template = template
        self.request = request
        self.object = object
        self.data = data
        self.context: dict = {
            "request": request,
            **(context if context is not None else {}),
        }

        self._valid = True
        self._form: F | None = None

    async def get_form(self) -> F:
        if self._form:
            return self._form

        formdata = None
        if self.request.method in {"POST", "PUT", "PATCH"}:
            formdata = await self.request.form()
        self._form = self.form_class(
            formdata=formdata,
            obj=self.object,
            data=self.data,
            meta={"request": self.request},
        )
        self.context.update({"form": self._form})
        return self._form

    async def is_submitted_and_valid(self) -> bool:
        self._form = await self.get_form()
        if self.request.method in {"POST", "PUT", "PATCH"}:
            self._valid = self._form.validate()
            return self._valid
        return False

    async def get_response(self) -> _TemplateResponse:
        await self.get_form()
        status_code = status.HTTP_200_OK if self._valid else status.HTTP_400_BAD_REQUEST
        return templates.TemplateResponse(
            self.template, self.context, status_code=status_code
        )

    async def get_error_response(
        self,
        error: str,
        error_code: str,
        *,
        fatal: bool = False,
        status_code: int = status.HTTP_400_BAD_REQUEST,
    ) -> _TemplateResponse:
        self.context.update({"error": error, "fatal_error": fatal})
        return templates.TemplateResponse(
            self.template,
            self.context,
            status_code=status_code,
            headers={"X-Fief-Error": error_code},
        )


class ComboboxSelectField(HiddenField):
    def __init__(
        self,
        *args,
        query_endpoint_path: str,
        query_parameter_name: str = "query",
        value_attr: str = "id",
        label_attr: str = "name",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.query_endpoint_path = query_endpoint_path
        self.query_parameter_name = query_parameter_name
        self.value_attr = value_attr
        self.label_attr = label_attr

    def _value(self):
        if self.data is not None:
            try:
                return getattr(self.data, self.value_attr)
            except AttributeError:
                pass
        return ""


class ComboboxSelectMultipleField(SelectMultipleField):
    def __init__(
        self,
        *args,
        query_endpoint_path: str,
        query_parameter_name: str = "query",
        value_attr: str = "id",
        label_attr: str = "name",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.query_endpoint_path = query_endpoint_path
        self.query_parameter_name = query_parameter_name
        self.value_attr = value_attr
        self.label_attr = label_attr

    def _choices_generator(self, choices):
        if choices:
            if isinstance(choices[0], (list, tuple)):
                _choices = choices
            else:
                _choices = zip(choices, choices)
        else:
            _choices = []

        data_values = []
        for item in self.data or []:
            try:
                data_values.append(getattr(item, self.value_attr))
            except AttributeError:
                pass

        for value, label in _choices:
            selected = value in data_values
            yield (value, label, selected)

    def process_data(self, value):
        try:
            self.data = list(value)
        except (ValueError, TypeError):
            self.data = None


class SelectMultipleFieldCheckbox(SelectMultipleField):
    widget = widgets.ListWidget()
    option_widget = widgets.CheckboxInput()


class TimezoneField(SelectField):
    def __init__(self, *args, **kwargs):
        choices = [""] + sorted(pytz.common_timezones)
        super().__init__(*args, choices=choices, **kwargs)


def empty_string_to_none(value: str | None) -> str | None:
    if value is None or value == "":
        return None
    return value


class PhoneNumberField(TelField):
    def process_data(self, value):
        if value is None or value is unset_value:
            self.data = None
            return

        self.data = self._validate_phone_number(value)

    def process_formdata(self, valuelist):
        if not valuelist:
            return

        self.data = self._validate_phone_number(valuelist[0])

    def _validate_phone_number(self, value: str) -> str:
        try:
            parsed = phonenumbers.parse(value)
        except phonenumbers.phonenumberutil.NumberParseException as e:
            raise ValueError(_("The country code is missing.")) from e
        if not phonenumbers.is_valid_number(parsed):
            raise ValueError(_("This phone number is invalid."))
        return phonenumbers.format_number(parsed, phonenumbers.PhoneNumberFormat.E164)


class CountryField(SelectField):
    def __init__(self, *args, **kwargs):
        countries = sorted(pycountry.countries, key=lambda c: c.name)
        choices = [("", "")] + [
            (country.alpha_2, country.name) for country in countries
        ]
        super().__init__(*args, choices=choices, **kwargs)


class AddressForm(BaseForm):
    def __init__(self, *args, required: bool = True, **kwargs):
        self.required = required
        super().__init__(*args, **kwargs)

    line1 = StringField(_("Address line 1"), validators=[validators.DataRequired()])
    line2 = StringField(
        _("Address line 2"),
        validators=[validators.Optional()],
        filters=[empty_string_to_none],
    )
    postal_code = StringField(_("Postal code"), validators=[validators.DataRequired()])
    city = StringField(_("City"), validators=[validators.DataRequired()])
    state = StringField(
        _("State"), validators=[validators.Optional()], filters=[empty_string_to_none]
    )
    country = CountryField(_("Country"), validators=[validators.DataRequired()])

    def validate(self, extra_validators=None):
        if self.data is None and not self.required:
            return True
        return super().validate(extra_validators)

    @property
    def data(self):
        data = super().data
        if not any(data.values()):
            return None
        return data


class AddressFormField(FormField):
    def __init__(self, *args, required: bool = True, **kwargs):
        form_class = functools.partial(AddressForm, required=required)
        super().__init__(form_class, separator=".", *args, **kwargs)


class PasswordFieldForm(BaseForm):
    password = PasswordField(_("Password"), validators=[validators.InputRequired()])


USER_FIELD_FORM_FIELD_MAP: Mapping[UserFieldType, Field] = {
    UserFieldType.STRING: StringField,
    UserFieldType.INTEGER: IntegerField,
    UserFieldType.BOOLEAN: BooleanField,
    UserFieldType.DATE: DateField,
    UserFieldType.DATETIME: DateTimeLocalField,
    UserFieldType.CHOICE: SelectField,
    UserFieldType.PHONE_NUMBER: PhoneNumberField,
    UserFieldType.ADDRESS: AddressFormField,
    UserFieldType.TIMEZONE: TimezoneField,
}


def get_form_field(user_field: UserField) -> Field:
    field_validators = []
    required = user_field.get_required()
    if required:
        field_validators.append(validators.InputRequired())
    else:
        field_validators.append(validators.Optional())

    field_kwargs = {
        "label": user_field.name,
        "validators": field_validators,
        "default": user_field.configuration.get("default"),
    }

    if user_field.type == UserFieldType.STRING:
        field_kwargs.update({"filters": [empty_string_to_none]})
    if user_field.type == UserFieldType.CHOICE:
        field_kwargs.update({"choices": user_field.configuration.get("choices")})
    elif user_field.type == UserFieldType.PHONE_NUMBER:
        field_kwargs.update({"render_kw": {"placeholder": "+42102030405"}})
    elif user_field.type == UserFieldType.ADDRESS:
        field_kwargs.update({"required": required})
        field_kwargs.pop("validators")

    return USER_FIELD_FORM_FIELD_MAP[user_field.type](**field_kwargs)



from datetime import datetime, timezone

from sqlalchemy import select

from fief.models import LoginSession
from fief.repositories.base import BaseRepository, ExpiresAtMixin, UUIDRepositoryMixin


class LoginSessionRepository(
    BaseRepository[LoginSession],
    UUIDRepositoryMixin[LoginSession],
    ExpiresAtMixin[LoginSession],
):
    model = LoginSession

    async def get_by_token(
        self, token: str, *, fresh: bool = True
    ) -> LoginSession | None:
        statement = select(LoginSession).where(LoginSession.token == token)
        if fresh:
            statement = statement.where(
                LoginSession.expires_at > datetime.now(timezone.utc)
            )
        return await self.get_one_or_none(statement)



from datetime import datetime

from pydantic import UUID4
from sqlalchemy import JSON, ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fief.models.base import WorkspaceBase
from fief.models.client import Client
from fief.models.generics import (
    GUID,
    CreatedUpdatedAt,
    ExpiresAt,
    TIMESTAMPAware,
    UUIDModel,
)
from fief.models.user import User


class RefreshToken(UUIDModel, CreatedUpdatedAt, ExpiresAt, WorkspaceBase):
    __tablename__ = "refresh_tokens"

    token: Mapped[str] = mapped_column(
        String(length=255),
        nullable=False,
        index=True,
        unique=True,
    )
    scope: Mapped[list[str]] = mapped_column(JSON, nullable=False, default=list)
    authenticated_at: Mapped[datetime] = mapped_column(
        TIMESTAMPAware(timezone=True), nullable=False
    )

    user_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False
    )
    user: Mapped[User] = relationship("User")

    client_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Client.id, ondelete="CASCADE"), nullable=False
    )
    client: Mapped[Client] = relationship("Client", lazy="joined")



from fastapi.security.oauth2 import OAuth2PasswordRequestForm
from wtforms import EmailField, PasswordField, SubmitField, validators

from fief.forms import CSRFBaseForm
from fief.locale import gettext_lazy as _


class LoginForm(CSRFBaseForm):
    email = EmailField(
        _("Email address"), validators=[validators.InputRequired(), validators.Email()]
    )
    password = PasswordField(_("Password"), validators=[validators.InputRequired()])

    def get_credentials(self) -> OAuth2PasswordRequestForm:
        return OAuth2PasswordRequestForm(
            username=self.email.data,
            password=self.password.data,
            grant_type="password",
            scope="",
        )


class ConsentForm(CSRFBaseForm):
    allow = SubmitField(_("Allow"))
    deny = SubmitField(_("Deny"))



from fastapi import APIRouter, Depends, HTTPException, Request, Response, status
from fastapi.responses import JSONResponse
from fastapi_users.exceptions import InvalidPasswordException, UserAlreadyExists
from pydantic import UUID4
from sqlalchemy.orm import joinedload

from fief import schemas, tasks
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.tasks import get_send_task
from fief.dependencies.user_field import get_user_fields
from fief.dependencies.users import (
    UserManager,
    get_admin_user_update,
    get_paginated_user_oauth_accounts,
    get_paginated_user_permissions,
    get_paginated_user_roles,
    get_paginated_users,
    get_user_by_id_or_404,
    get_user_create_internal,
    get_user_manager_from_create_user_internal,
    get_user_manager_from_user,
)
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.errors import APIErrorCode
from fief.logger import AuditLogger
from fief.models import (
    AuditLogMessage,
    OAuthAccount,
    User,
    UserField,
    UserPermission,
    UserRole,
    Workspace,
)
from fief.repositories import (
    PermissionRepository,
    RoleRepository,
    UserPermissionRepository,
    UserRepository,
    UserRoleRepository,
)
from fief.schemas.generics import PaginatedResults
from fief.services.webhooks.models import (
    UserDeleted,
    UserPermissionCreated,
    UserPermissionDeleted,
    UserRoleCreated,
    UserRoleDeleted,
)
from fief.tasks import SendTask

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/", name="users:list", response_model=PaginatedResults[schemas.user.UserRead]
)
async def list_users(
    paginated_users: PaginatedObjects[User] = Depends(get_paginated_users),
) -> PaginatedResults[schemas.user.UserRead]:
    users, count = paginated_users
    return PaginatedResults(
        count=count,
        results=[schemas.user.UserRead.from_orm(user) for user in users],
    )


@router.get("/{id:uuid}", name="users:get", response_model=schemas.user.UserRead)
async def get_user(
    user: User = Depends(get_user_by_id_or_404),
) -> schemas.user.UserRead:
    return schemas.user.UserRead.from_orm(user)


@router.post(
    "/",
    name="users:create",
    response_model=schemas.user.UserRead,
    status_code=status.HTTP_201_CREATED,
)
async def create_user(
    request: Request,
    user_create: schemas.user.UserCreateInternal = Depends(get_user_create_internal),
    user_fields: list[UserField] = Depends(get_user_fields),
    user_manager: UserManager = Depends(get_user_manager_from_create_user_internal),
    user_repository: UserRepository = Depends(get_workspace_repository(UserRepository)),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    try:
        created_user = await user_manager.create_with_fields(
            user_create, user_fields=user_fields, request=request
        )
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, created_user)
    except UserAlreadyExists as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_CREATE_ALREADY_EXISTS,
        ) from e
    except InvalidPasswordException as e:
        # Build a JSON response manually to fine-tune the response structure
        return JSONResponse(
            content={
                "detail": APIErrorCode.USER_CREATE_INVALID_PASSWORD,
                "reason": str(e.reason),
            },
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    user = await user_repository.get_by_id(created_user.id, (joinedload(User.tenant),))

    return schemas.user.UserRead.from_orm(user)


@router.patch("/{id:uuid}", name="users:update", response_model=schemas.user.UserRead)
async def update_user(
    request: Request,
    user_update: schemas.user.UserUpdate = Depends(get_admin_user_update),
    user: User = Depends(get_user_by_id_or_404),
    user_fields: list[UserField] = Depends(get_user_fields),
    user_manager: UserManager = Depends(get_user_manager_from_user),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    try:
        user = await user_manager.update_with_fields(
            user_update,
            user,
            user_fields=user_fields,
            safe=False,
            request=request,
        )
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, user)
    except UserAlreadyExists as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_UPDATE_EMAIL_ALREADY_EXISTS,
        ) from e
    except InvalidPasswordException as e:
        # Build a JSON response manually to fine-tune the response structure
        return JSONResponse(
            content={
                "detail": APIErrorCode.USER_UPDATE_INVALID_PASSWORD,
                "reason": str(e.reason),
            },
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    return schemas.user.UserRead.from_orm(user)


@router.delete(
    "/{id:uuid}",
    name="users:delete",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_user(
    user: User = Depends(get_user_by_id_or_404),
    repository: UserRepository = Depends(get_workspace_repository(UserRepository)),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    await repository.delete(user)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, user)
    trigger_webhooks(UserDeleted, user, schemas.user.UserRead)


@router.get(
    "/{id:uuid}/permissions",
    name="users:list_permissions",
    response_model=PaginatedResults[schemas.user_permission.UserPermission],
)
async def list_user_permissions(
    paginated_user_permissions: PaginatedObjects[UserPermission] = Depends(
        get_paginated_user_permissions
    ),
) -> PaginatedResults[schemas.user_permission.UserPermission]:
    user_permissions, count = paginated_user_permissions
    return PaginatedResults(
        count=count,
        results=[
            schemas.user_permission.UserPermission.from_orm(user_permission)
            for user_permission in user_permissions
        ],
    )


@router.post(
    "/{id:uuid}/permissions",
    name="users:create_permission",
    status_code=status.HTTP_201_CREATED,
)
async def create_user_permission(
    user_permission_create: schemas.user_permission.UserPermissionCreate,
    user: User = Depends(get_user_by_id_or_404),
    permission_repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    user_permission_repository: UserPermissionRepository = Depends(
        get_workspace_repository(UserPermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> None:
    permission_id = user_permission_create.id
    permission = await permission_repository.get_by_id(permission_id)

    if permission is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_PERMISSION_CREATE_NOT_EXISTING_PERMISSION,
        )

    existing_user_permission = (
        await user_permission_repository.get_by_permission_and_user(
            user.id, permission_id, direct_only=True
        )
    )
    if existing_user_permission is not None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_PERMISSION_CREATE_ALREADY_ADDED_PERMISSION,
        )

    user_permission = UserPermission(user_id=user.id, permission=permission)
    await user_permission_repository.create(user_permission)
    audit_logger.log_object_write(
        AuditLogMessage.OBJECT_CREATED,
        user_permission,
        subject_user_id=user.id,
        permission_id=str(permission.id),
    )
    trigger_webhooks(
        UserPermissionCreated,
        user_permission,
        schemas.user_permission.UserPermission,
    )


@router.delete(
    "/{id:uuid}/permissions/{permission_id:uuid}",
    name="users:delete_permission",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_user_permission(
    permission_id: UUID4,
    user: User = Depends(get_user_by_id_or_404),
    user_permission_repository: UserPermissionRepository = Depends(
        get_workspace_repository(UserPermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> None:
    user_permission = await user_permission_repository.get_by_permission_and_user(
        user.id, permission_id, direct_only=True
    )
    if user_permission is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    await user_permission_repository.delete(user_permission)
    audit_logger.log_object_write(
        AuditLogMessage.OBJECT_DELETED,
        user_permission,
        subject_user_id=user.id,
        permission_id=str(permission_id),
    )
    trigger_webhooks(
        UserPermissionDeleted,
        user_permission,
        schemas.user_permission.UserPermission,
    )


@router.get(
    "/{id:uuid}/roles",
    name="users:list_roles",
    response_model=PaginatedResults[schemas.user_role.UserRole],
)
async def list_user_roles(
    paginated_user_roles: PaginatedObjects[UserRole] = Depends(
        get_paginated_user_roles
    ),
) -> PaginatedResults[schemas.user_role.UserRole]:
    user_roles, count = paginated_user_roles
    return PaginatedResults(
        count=count,
        results=[
            schemas.user_role.UserRole.from_orm(user_role) for user_role in user_roles
        ],
    )


@router.post(
    "/{id:uuid}/roles", name="users:create_role", status_code=status.HTTP_201_CREATED
)
async def create_user_role(
    user_role_create: schemas.user_role.UserRoleCreate,
    user: User = Depends(get_user_by_id_or_404),
    role_repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    user_role_repository: UserRoleRepository = Depends(
        get_workspace_repository(UserRoleRepository)
    ),
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> None:
    role_id = user_role_create.id
    role = await role_repository.get_by_id(role_id)

    if role is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_ROLE_CREATE_NOT_EXISTING_ROLE,
        )

    existing_user_role = await user_role_repository.get_by_role_and_user(
        user.id, role_id
    )
    if existing_user_role is not None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_ROLE_CREATE_ALREADY_ADDED_ROLE,
        )

    user_role = UserRole(user_id=user.id, role=role)
    await user_role_repository.create(user_role)
    audit_logger.log_object_write(
        AuditLogMessage.OBJECT_CREATED,
        user_role,
        subject_user_id=user.id,
        role_id=str(role.id),
    )
    trigger_webhooks(UserRoleCreated, user_role, schemas.user_role.UserRole)

    send_task(tasks.on_user_role_created, str(user.id), str(role.id), str(workspace.id))


@router.delete(
    "/{id:uuid}/roles/{role_id:uuid}",
    name="users:delete_role",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_user_role(
    role_id: UUID4,
    user: User = Depends(get_user_by_id_or_404),
    user_role_repository: UserRoleRepository = Depends(
        get_workspace_repository(UserRoleRepository)
    ),
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> None:
    user_role = await user_role_repository.get_by_role_and_user(user.id, role_id)
    if user_role is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    await user_role_repository.delete(user_role)
    audit_logger.log_object_write(
        AuditLogMessage.OBJECT_DELETED,
        user_role,
        subject_user_id=user.id,
        role_id=str(role_id),
    )
    trigger_webhooks(UserRoleDeleted, user_role, schemas.user_role.UserRole)

    send_task(tasks.on_user_role_deleted, str(user.id), str(role_id), str(workspace.id))


@router.get(
    "/{id:uuid}/oauth-accounts",
    name="users:list_oauth_accounts",
    response_model=PaginatedResults[schemas.oauth_account.OAuthAccount],
)
async def list_user_oauth_accounts(
    paginated_user_oauth_accounts: PaginatedObjects[OAuthAccount] = Depends(
        get_paginated_user_oauth_accounts
    ),
) -> PaginatedResults[schemas.oauth_account.OAuthAccount]:
    oauth_accounts, count = paginated_user_oauth_accounts
    return PaginatedResults(
        count=count,
        results=[
            schemas.oauth_account.OAuthAccount.from_orm(oauth_account)
            for oauth_account in oauth_accounts
        ],
    )



from fastapi import Depends, HTTPException, Query, status
from pydantic import UUID4
from sqlalchemy import select

from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import Client
from fief.repositories import ClientRepository


async def get_paginated_clients(
    query: str | None = Query(None),
    tenant: UUID4 | None = Query(None),
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    get_paginated_objects: GetPaginatedObjects[Client] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[Client]:
    statement = select(Client)

    if query is not None:
        statement = statement.where(Client.name.ilike(f"%{query}%"))

    if tenant is not None:
        statement = statement.where(Client.tenant_id == tenant)

    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_client_by_id_or_404(
    id: UUID4,
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
) -> Client:
    client = await repository.get_by_id(id)

    if client is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return client



import uuid

import dramatiq

from fief import schemas
from fief.services.email_template.contexts import ForgotPasswordContext
from fief.services.email_template.types import EmailTemplateType
from fief.tasks.base import TaskBase


class OnAfterForgotPasswordTask(TaskBase):
    __name__ = "on_after_forgot_password"

    async def run(self, user_id: str, workspace_id: str, reset_url: str):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))
        user = await self._get_user(uuid.UUID(user_id), workspace)
        tenant = await self._get_tenant(user.tenant_id, workspace)

        context = ForgotPasswordContext(
            tenant=schemas.tenant.Tenant.from_orm(tenant),
            user=schemas.user.UserEmailContext.from_orm(user),
            reset_url=reset_url,
        )

        async with self._get_email_subject_renderer(
            workspace
        ) as email_subject_renderer:
            subject = await email_subject_renderer.render(
                EmailTemplateType.FORGOT_PASSWORD, context
            )

        async with self._get_email_template_renderer(
            workspace
        ) as email_template_renderer:
            html = await email_template_renderer.render(
                EmailTemplateType.FORGOT_PASSWORD, context
            )

        self.email_provider.send_email(
            sender=("contact@fief.dev", None),
            recipient=(user.email, None),
            subject=subject,
            html=html,
        )


on_after_forgot_password = dramatiq.actor(OnAfterForgotPasswordTask())



from fastapi import APIRouter, Request
from fastapi.param_functions import Depends
from jwcrypto import jwk
from starlette.routing import Router

from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.tenant import get_current_tenant
from fief.models import Tenant, Workspace
from fief.schemas.well_known import OpenIDProviderMetadata
from fief.services.response_type import ALLOWED_RESPONSE_TYPES
from fief.settings import settings

router = APIRouter()


@router.get("/openid-configuration", name="well_known:openid_configuration")
async def get_openid_configuration(
    request: Request,
    workspace: Workspace = Depends(get_current_workspace),
    tenant: Tenant = Depends(get_current_tenant),
):
    url_for_params = {}
    if not tenant.default:
        url_for_params["tenant_slug"] = tenant.slug

    def _url_for(name: str) -> str:
        router: Router = request.scope["router"]
        url_path = router.url_path_for(name, **url_for_params)

        x_forwarded_host = request.headers.get("X-Forwarded-Host", None)
        host = x_forwarded_host if x_forwarded_host else request.base_url.netloc
        base_url = request.base_url.replace(netloc=host)

        return url_path.make_absolute_url(base_url)

    configuration = OpenIDProviderMetadata(
        issuer=tenant.get_host(workspace.domain),
        authorization_endpoint=_url_for("auth:authorize"),
        token_endpoint=_url_for("auth:token"),
        userinfo_endpoint=_url_for("user:userinfo.get"),
        jwks_uri=_url_for("well_known:jwks"),
        registration_endpoint=_url_for("register:register"),
        scopes_supported=["openid", "offline_access"],
        response_types_supported=ALLOWED_RESPONSE_TYPES,
        response_modes_supported=["query", "fragment"],
        grant_types_supported=["authorization_code", "refresh_token"],
        subject_types_supported=["public"],
        id_token_signing_alg_values_supported=["RS256"],
        id_token_encryption_alg_values_supported=["RSA-OAEP-256"],
        id_token_encryption_enc_values_supported=["A256CBC-HS512"],
        userinfo_signing_alg_values_supported=["none"],
        token_endpoint_auth_methods_supported=[
            "client_secret_basic",
            "client_secret_post",
        ],
        claims_supported=["email", "tenant_id"],
        request_parameter_supported=False,
        code_challenge_methods_supported=["plain", "S256"],
        service_documentation=settings.fief_documentation_url,
    )

    return configuration.dict(exclude_unset=True)


@router.get("/jwks.json", name="well_known:jwks")
async def get_jwks(tenant: Tenant = Depends(get_current_tenant)):
    keyset = jwk.JWKSet(keys=tenant.get_sign_jwk())
    return keyset.export(private_keys=False, as_dict=True)



from collections.abc import Callable

from fastapi import Request
from starlette.exceptions import HTTPException as StarletteHTTPException

from fief.templates import templates

exception_handlers: dict[type[Exception], Callable] = {}


async def http_exception_handler(request: Request, exc: StarletteHTTPException):
    headers = getattr(exc, "headers", None)
    return templates.TemplateResponse(
        "admin/error.html",
        {
            "request": request,
            "status_code": exc.status_code,
            "detail": exc.detail,
        },
        status_code=exc.status_code,
        headers=headers,
    )


exception_handlers[StarletteHTTPException] = http_exception_handler


__all__ = ["exception_handlers"]



import json
from typing import Any

from fastapi import Request
from fastapi.exceptions import RequestValidationError
from pydantic.error_wrappers import ErrorWrapper


async def get_request_json(request: Request) -> dict[str, Any]:
    try:
        return await request.json()
    except json.JSONDecodeError as e:
        # Taken from FastAPI to mimic its builtin logic when encountering invalid JSON
        # Ref: https://github.com/tiangolo/fastapi/blob/7b3727e03e84ca202d450ba3d702d5cd37025d60/fastapi/routing.py#L217-L220
        raise RequestValidationError(
            [ErrorWrapper(e, ("body", e.pos))], body=e.doc
        ) from e



from datetime import datetime, timezone

from fastapi import APIRouter, Depends, Query, Request, status
from fastapi.responses import RedirectResponse
from httpx_oauth.errors import GetIdEmailError
from httpx_oauth.oauth2 import GetAccessTokenError

from fief.dependencies.authentication_flow import get_authentication_flow
from fief.dependencies.oauth import (
    get_login_session_with_tenant_query,
    get_oauth_provider,
    get_tenant_by_query,
)
from fief.dependencies.oauth_provider import get_oauth_providers
from fief.dependencies.register import get_registration_flow
from fief.dependencies.session_token import get_session_token
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.exceptions import OAuthException
from fief.locale import gettext_lazy as _
from fief.models import (
    LoginSession,
    OAuthAccount,
    OAuthProvider,
    OAuthSession,
    RegistrationSessionFlow,
    SessionToken,
    Tenant,
)
from fief.repositories import OAuthAccountRepository, OAuthSessionRepository
from fief.schemas.oauth import OAuthError
from fief.services.authentication_flow import AuthenticationFlow
from fief.services.oauth_provider import get_oauth_provider_service
from fief.services.registration_flow import RegistrationFlow

router = APIRouter(prefix="/oauth")


@router.get("/authorize", name="oauth:authorize")
async def authorize(
    request: Request,
    tenant: Tenant = Depends(get_tenant_by_query),
    login_session: LoginSession = Depends(get_login_session_with_tenant_query),
    oauth_provider: OAuthProvider = Depends(get_oauth_provider),
    oauth_session_repository: OAuthSessionRepository = Depends(
        get_workspace_repository(OAuthSessionRepository)
    ),
):
    redirect_uri = request.url_for("oauth:callback")

    oauth_session = OAuthSession(
        redirect_uri=redirect_uri,
        oauth_provider=oauth_provider,
        login_session=login_session,
        tenant=tenant,
    )
    oauth_session = await oauth_session_repository.create(oauth_session)
    state = oauth_session.token

    oauth_provider_service = get_oauth_provider_service(oauth_provider)
    scopes = set(oauth_provider_service.base_scopes or []) | set(oauth_provider.scopes)
    authorize_url = await oauth_provider_service.get_authorization_url(
        redirect_uri, state=state, scope=list(scopes)
    )

    return RedirectResponse(authorize_url, status_code=status.HTTP_302_FOUND)


@router.get("/callback", name="oauth:callback")
async def callback(
    request: Request,
    code: str | None = Query(None),
    code_verifier: str | None = Query(None),
    state: str | None = Query(None),
    error: str | None = Query(None),
    oauth_providers: list[OAuthProvider] | None = Depends(get_oauth_providers),
    oauth_session_repository: OAuthSessionRepository = Depends(
        get_workspace_repository(OAuthSessionRepository)
    ),
    oauth_account_repository: OAuthAccountRepository = Depends(
        get_workspace_repository(OAuthAccountRepository)
    ),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    registration_flow: RegistrationFlow = Depends(get_registration_flow),
    session_token: SessionToken | None = Depends(get_session_token),
):
    if error is not None:
        raise OAuthException(
            OAuthError.get_oauth_error(error),
            oauth_providers=oauth_providers,
            fatal=True,
        )

    if code is None:
        raise OAuthException(
            OAuthError.get_missing_code(_("Missing authorization code.")),
            oauth_providers=oauth_providers,
            fatal=True,
        )

    oauth_session = (
        await oauth_session_repository.get_by_token(state)
        if state is not None
        else None
    )
    if oauth_session is None:
        raise OAuthException(
            OAuthError.get_invalid_session(_("Invalid OAuth session.")),
            oauth_providers=oauth_providers,
            fatal=True,
        )

    tenant = oauth_session.tenant
    oauth_provider = oauth_session.oauth_provider
    oauth_provider_service = get_oauth_provider_service(oauth_provider)

    try:
        access_token_dict = await oauth_provider_service.get_access_token(
            code, oauth_session.redirect_uri, code_verifier
        )
    except GetAccessTokenError as e:
        raise OAuthException(
            OAuthError.get_access_token_error(
                _("An error occurred while getting the access token.")
            ),
            oauth_providers=oauth_providers,
            tenant=tenant,
        ) from e

    access_token = access_token_dict["access_token"]
    refresh_token = access_token_dict.get("refresh_token")
    try:
        expires_at = datetime.fromtimestamp(
            access_token_dict["expires_at"], tz=timezone.utc
        )
    except KeyError:
        expires_at = None

    try:
        account_id, account_email = await oauth_provider_service.get_id_email(
            access_token
        )
    except GetIdEmailError as e:
        raise OAuthException(
            OAuthError.get_id_email_error(
                _(
                    "An error occurred while querying the provider API. Original error message: %(message)s",
                    message=str(e),
                )
            ),
            oauth_providers=oauth_providers,
            tenant=tenant,
        ) from e

    oauth_account = await oauth_account_repository.get_by_provider_and_account_id(
        oauth_provider.id, account_id
    )

    # Existing account
    if oauth_account is not None and oauth_account.user is not None:
        user = oauth_account.user
        if not user.is_active:
            raise OAuthException(
                OAuthError.get_inactive_user(_("Your account is inactive.")),
                oauth_providers=oauth_providers,
                tenant=tenant,
            )

        # Update tokens
        oauth_account.access_token = access_token
        oauth_account.refresh_token = refresh_token
        oauth_account.expires_at = expires_at
        await oauth_account_repository.update(oauth_account)

        # Redirect to consent
        response = RedirectResponse(
            tenant.url_path_for(request, "auth:consent"),
            status_code=status.HTTP_302_FOUND,
        )
        response = await authentication_flow.rotate_session_token(
            response, user.id, session_token=session_token
        )
        return response

    # New account to create
    oauth_account = OAuthAccount(
        access_token=access_token,
        expires_at=expires_at,
        refresh_token=refresh_token,
        account_id=account_id,
        account_email=account_email,
        oauth_provider=oauth_provider,
        tenant_id=tenant.id,
    )
    oauth_account = await oauth_account_repository.create(oauth_account)
    oauth_session.oauth_account = oauth_account
    await oauth_session_repository.update(oauth_session)

    response = RedirectResponse(
        tenant.url_path_for(request, "register:register"),
        status_code=status.HTTP_302_FOUND,
    )

    await registration_flow.create_registration_session(
        response,
        RegistrationSessionFlow.OAUTH,
        tenant=tenant,
        oauth_account=oauth_account,
    )

    return response



from fastapi import APIRouter, Depends, Query, Request, status
from fastapi.responses import RedirectResponse
from pydantic import AnyUrl

from fief.apps.auth.forms.auth import ConsentForm, LoginForm
from fief.dependencies.auth import (
    check_unsupported_request_parameter,
    get_authorize_client,
    get_authorize_code_challenge,
    get_authorize_prompt,
    get_authorize_redirect_uri,
    get_authorize_response_mode,
    get_authorize_response_type,
    get_authorize_scope,
    get_authorize_screen,
    get_consent_prompt,
    get_login_session,
    get_needs_consent,
    get_nonce,
    has_valid_session_token,
)
from fief.dependencies.authentication_flow import get_authentication_flow
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.oauth_provider import get_oauth_providers
from fief.dependencies.session_token import get_session_token
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.theme import get_current_theme
from fief.dependencies.users import UserManager, get_user_manager
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.exceptions import LogoutException
from fief.forms import FormHelper
from fief.locale import gettext_lazy as _
from fief.models import Client, LoginSession, OAuthProvider, Tenant, Theme, Workspace
from fief.models.session_token import SessionToken
from fief.repositories.session_token import SessionTokenRepository
from fief.schemas.auth import LogoutError
from fief.services.authentication_flow import AuthenticationFlow
from fief.settings import settings

router = APIRouter()


@router.get(
    "/authorize",
    name="auth:authorize",
    dependencies=[Depends(check_unsupported_request_parameter)],
)
async def authorize(
    request: Request,
    response_type: str = Depends(get_authorize_response_type),
    client: Client = Depends(get_authorize_client),
    redirect_uri: str = Depends(get_authorize_redirect_uri),
    response_mode: str = Depends(get_authorize_response_mode),
    scope: list[str] = Depends(get_authorize_scope),
    prompt: str | None = Depends(get_authorize_prompt),
    screen: str = Depends(get_authorize_screen),
    code_challenge_tuple: tuple[str, str]
    | None = Depends(get_authorize_code_challenge),
    nonce: str | None = Depends(get_nonce),
    state: str | None = Query(None),
    lang: str | None = Query(None),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    has_valid_session_token: bool = Depends(has_valid_session_token),
):
    tenant = client.tenant

    if has_valid_session_token and prompt != "login":
        redirection = tenant.url_path_for(request, "auth:consent")
    elif screen == "register":
        redirection = tenant.url_path_for(request, "register:register")
    else:
        redirection = tenant.url_path_for(request, "auth:login")

    response = RedirectResponse(url=redirection, status_code=status.HTTP_302_FOUND)
    response = await authentication_flow.create_login_session(
        response,
        response_type=response_type,
        response_mode=response_mode,
        redirect_uri=redirect_uri,
        scope=scope,
        state=state,
        nonce=nonce,
        code_challenge_tuple=code_challenge_tuple,
        client=client,
    )

    if lang is not None:
        response.set_cookie(
            settings.user_locale_cookie_name,
            lang,
            max_age=settings.user_locale_lifetime_seconds,
            domain=settings.session_cookie_domain,
            secure=settings.session_cookie_secure,
            httponly=True,
        )

    return response


@router.api_route(
    "/login",
    methods=["GET", "POST"],
    name="auth:login",
    dependencies=[Depends(get_login_session)],
)
async def login(
    request: Request,
    user_manager: UserManager = Depends(get_user_manager),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    session_token: SessionToken | None = Depends(get_session_token),
    oauth_providers: list[OAuthProvider] | None = Depends(get_oauth_providers),
    tenant: Tenant = Depends(get_current_tenant),
    theme: Theme = Depends(get_current_theme),
):
    form_helper = FormHelper(
        LoginForm,
        "auth/login.html",
        request=request,
        context={"oauth_providers": oauth_providers, "tenant": tenant, "theme": theme},
    )
    form = await form_helper.get_form()

    if await form_helper.is_submitted_and_valid():
        user = await user_manager.authenticate(form.get_credentials())
        if user is None or not user.is_active:
            return await form_helper.get_error_response(
                _("Invalid email or password"), "bad_credentials"
            )

        response = RedirectResponse(
            tenant.url_path_for(request, "auth:consent"),
            status_code=status.HTTP_302_FOUND,
        )
        response = await authentication_flow.rotate_session_token(
            response, user.id, session_token=session_token
        )
        return response

    return await form_helper.get_response()


@router.api_route("/consent", methods=["GET", "POST"], name="auth:consent")
async def consent(
    request: Request,
    login_session: LoginSession = Depends(get_login_session),
    session_token: SessionToken | None = Depends(get_session_token),
    prompt: str | None = Depends(get_consent_prompt),
    needs_consent: bool = Depends(get_needs_consent),
    tenant: Tenant = Depends(get_current_tenant),
    theme: Theme = Depends(get_current_theme),
    workspace: Workspace = Depends(get_current_workspace),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
):
    form_helper = FormHelper(
        ConsentForm,
        "auth/consent.html",
        request=request,
        context={
            "tenant": tenant,
            "theme": theme,
            "client": login_session.client,
            "scopes": login_session.scope,
        },
    )
    form = await form_helper.get_form()

    if session_token is None:
        return RedirectResponse(
            tenant.url_path_for(request, "auth:login"),
            status_code=status.HTTP_302_FOUND,
        )

    if not needs_consent and prompt != "consent":
        response = await authentication_flow.get_authorization_code_success_redirect(
            login_session=login_session,
            authenticated_at=session_token.created_at,
            user=session_token.user,
            client=login_session.client,
            tenant=tenant,
            workspace=workspace,
        )
        response = await authentication_flow.delete_login_session(
            response, login_session
        )
        return response

    if await form_helper.is_submitted_and_valid():
        # Allow
        if form.allow.data:
            user_id = session_token.user_id
            await authentication_flow.create_or_update_grant(
                user_id, login_session.client, login_session.scope
            )
            response = (
                await authentication_flow.get_authorization_code_success_redirect(
                    login_session=login_session,
                    authenticated_at=session_token.created_at,
                    user=session_token.user,
                    client=login_session.client,
                    tenant=tenant,
                    workspace=workspace,
                )
            )
        # Deny
        elif form.deny.data:
            response = AuthenticationFlow.get_authorization_code_error_redirect(
                login_session.redirect_uri,
                login_session.response_mode,
                "access_denied",
                error_description=_("The user denied access to their data."),
                state=login_session.state,
            )

        response = await authentication_flow.delete_login_session(
            response, login_session
        )
        return response

    return await form_helper.get_response()


@router.get("/logout", name="auth:logout")
async def logout(
    redirect_uri: AnyUrl | None = Query(None),
    session_token: SessionToken | None = Depends(get_session_token),
    sesstion_token_repository: SessionTokenRepository = Depends(
        get_workspace_repository(SessionTokenRepository)
    ),
    tenant: Tenant = Depends(get_current_tenant),
):
    if redirect_uri is None:
        raise LogoutException(
            LogoutError.get_invalid_request(_("redirect_uri is missing")),
            tenant,
        )

    if session_token is not None:
        await sesstion_token_repository.delete(session_token)

    response = RedirectResponse(redirect_uri, status_code=status.HTTP_302_FOUND)

    response.delete_cookie(
        settings.session_cookie_name,
        domain=settings.session_cookie_domain,
        secure=settings.session_cookie_secure,
        httponly=True,
    )

    return response



import uuid

import dramatiq

from fief.models import Role, UserPermission
from fief.repositories import (
    RoleRepository,
    UserPermissionRepository,
    UserRoleRepository,
)
from fief.tasks.base import ObjectDoesNotExistTaskError, TaskBase


class OnRoleUpdated(TaskBase):
    __name__ = "on_role_updated"

    async def run(
        self,
        role_id: str,
        added_permissions: set[str],
        deleted_permissions: set[str],
        workspace_id: str,
    ):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))

        async with self.get_workspace_session(workspace) as session:
            role_repository = RoleRepository(session)
            user_role_repository = UserRoleRepository(session)
            user_permission_repository = UserPermissionRepository(session)

            role = await role_repository.get_by_id(uuid.UUID(role_id))

            if role is None:
                raise ObjectDoesNotExistTaskError(Role, role_id)

            # Add newly added permissions to users with this role
            user_roles = await user_role_repository.get_by_role(role.id)
            user_permissions: list[UserPermission] = []
            for user_role in user_roles:
                for added_permission in added_permissions:
                    user_permissions.append(
                        UserPermission(
                            user_id=user_role.user_id,
                            permission_id=uuid.UUID(added_permission),
                            from_role_id=role.id,
                        )
                    )
            await user_permission_repository.create_many(user_permissions)

            # Revoke deleted permissions to users with this role
            for deleted_permission in deleted_permissions:
                await user_permission_repository.delete_by_permission_and_role(
                    uuid.UUID(deleted_permission), role.id
                )


on_role_updated = dramatiq.actor(OnRoleUpdated())



import ssl
from enum import StrEnum
from pathlib import Path
from typing import Union

from sqlalchemy import engine


class DatabaseType(StrEnum):
    POSTGRESQL = "POSTGRESQL"
    MYSQL = "MYSQL"
    SQLITE = "SQLITE"

    def get_display_name(self) -> str:
        display_names = {
            DatabaseType.POSTGRESQL: "PostgreSQL",
            DatabaseType.MYSQL: "MySQL",
            DatabaseType.SQLITE: "SQLite",
        }
        return display_names[self]

    @classmethod
    def choices(cls) -> list[tuple[str, str]]:
        return [
            (member.value, member.get_display_name())
            for member in cls
            if member != DatabaseType.SQLITE
        ]

    @classmethod
    def coerce(cls, item):
        return cls(str(item)) if not isinstance(item, cls) else item


class PostreSQLSSLMode(StrEnum):
    DISABLE = "disable"
    ALLOW = "allow"
    PREFER = "prefer"
    REQUIRE = "require"
    VERIFY_CA = "verify-ca"
    VERIFY_FULL = "verify-full"

    def get_display_name(self) -> str:
        display_names = {
            PostreSQLSSLMode.DISABLE: "Disabled",
            PostreSQLSSLMode.ALLOW: "Allow",
            PostreSQLSSLMode.PREFER: "Prefer",
            PostreSQLSSLMode.REQUIRE: "Require",
            PostreSQLSSLMode.VERIFY_CA: "Verify CA",
            PostreSQLSSLMode.VERIFY_FULL: "Verify Full",
        }
        return display_names[self]

    @classmethod
    def choices(cls) -> list[tuple[str, str]]:
        return [(member.value, member.get_display_name()) for member in cls]

    @classmethod
    def coerce(cls, item):
        return cls(str(item)) if not isinstance(item, cls) else item


class MySQLSSLMode(StrEnum):
    DISABLED = "DISABLED"
    PREFERRED = "PREFERRED"
    REQUIRED = "REQUIRED"
    VERIFY_CA = "VERIFY_CA"
    VERIFY_IDENTITY = "VERIFY_IDENTITY"

    def get_display_name(self) -> str:
        display_names = {
            MySQLSSLMode.DISABLED: "Disabled",
            MySQLSSLMode.PREFERRED: "Preferred",
            MySQLSSLMode.REQUIRED: "Required",
            MySQLSSLMode.VERIFY_CA: "Verify CA",
            MySQLSSLMode.VERIFY_IDENTITY: "Verify Identity",
        }
        return display_names[self]

    @classmethod
    def choices(cls) -> list[tuple[str, str]]:
        return [(member.value, member.get_display_name()) for member in cls]

    @classmethod
    def coerce(cls, item):
        return cls(str(item)) if not isinstance(item, cls) else item


SSLMode = Union[PostreSQLSSLMode, MySQLSSLMode]

SSL_MODES: dict[DatabaseType, type[SSLMode]] = {
    DatabaseType.POSTGRESQL: PostreSQLSSLMode,
    DatabaseType.MYSQL: MySQLSSLMode,
}

UNSAFE_SSL_MODES = {
    PostreSQLSSLMode.DISABLE,
    PostreSQLSSLMode.ALLOW,
    PostreSQLSSLMode.PREFER,
    MySQLSSLMode.DISABLED,
    MySQLSSLMode.PREFERRED,
}

SYNC_DRIVERS: dict[DatabaseType, str] = {
    DatabaseType.POSTGRESQL: "postgresql",
    DatabaseType.MYSQL: "mysql+pymysql",
    DatabaseType.SQLITE: "sqlite",
}

ASYNC_DRIVERS: dict[DatabaseType, str] = {
    DatabaseType.POSTGRESQL: "postgresql+asyncpg",
    DatabaseType.MYSQL: "mysql+aiomysql",
    DatabaseType.SQLITE: "sqlite+aiosqlite",
}


def get_driver(type: DatabaseType, *, asyncio: bool) -> str:
    drivers = ASYNC_DRIVERS if asyncio else SYNC_DRIVERS
    return drivers[type]


def get_ssl_mode_parameters(
    drivername: str, ssl_mode: str, query: dict[str, str], connect_args: dict
) -> tuple[dict[str, str], dict]:
    if ssl_mode in [PostreSQLSSLMode.DISABLE, MySQLSSLMode.DISABLED]:
        return query, connect_args

    # Build a SSL context
    context = ssl.create_default_context()

    # Basic mode: no hostname check, no certificate check
    context.check_hostname = False
    context.verify_mode = ssl.CERT_NONE

    # Verify CA mode, check the certificate
    if ssl_mode in [PostreSQLSSLMode.VERIFY_CA, MySQLSSLMode.VERIFY_CA]:
        context.verify_mode = ssl.CERT_REQUIRED

    # Verify full mode, check the certificate and hostname
    if ssl_mode in [PostreSQLSSLMode.VERIFY_FULL, MySQLSSLMode.VERIFY_IDENTITY]:
        context.check_hostname = True

    # PyMySQL does not support SSL context, it uses LibPQ directly so we just pass allowed query parameters
    if drivername == "postgresql":
        query["sslmode"] = ssl_mode
        query["sslrootcert"] = ssl.get_default_verify_paths().openssl_cafile
    elif drivername in ["postgresql+asyncpg", "mysql+aiomysql", "mysql+pymysql"]:
        connect_args["ssl"] = context

    return query, connect_args


DatabaseConnectionParameters = tuple[engine.URL, dict]


def create_database_connection_parameters(
    type: DatabaseType,
    *,
    asyncio: bool,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    path: Path | None = None,
    schema: str | None = None,
    ssl_mode: str | None = None,
) -> DatabaseConnectionParameters:
    drivername = get_driver(type, asyncio=asyncio)
    query: dict[str, str] = {}
    connect_args: dict = {}

    if ssl_mode:
        query, connect_args = get_ssl_mode_parameters(
            drivername, ssl_mode, query, connect_args
        )

    url = engine.URL.create(
        drivername=drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
        query=query,
    )

    dialect_name = url.get_dialect().name
    if dialect_name == "sqlite":
        name = schema if schema is not None else database
        assert name is not None
        assert path is not None
        url = url.set(database=str(path / name))

    return url, connect_args



"""Add logo_url column to Tenant

Revision ID: 31f6171a84fa
Revises: f72167c6b223
Create Date: 2023-01-06 08:26:27.744414

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "31f6171a84fa"
down_revision = "f72167c6b223"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "fief_tenants", sa.Column("logo_url", sa.String(length=512), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("fief_tenants", "logo_url")
    # ### end Alembic commands ###



from fief.models import Webhook
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class WebhookRepository(BaseRepository[Webhook], UUIDRepositoryMixin[Webhook]):
    model = Webhook



from sqlalchemy import select

from fief.models import Theme
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class ThemeRepository(BaseRepository[Theme], UUIDRepositoryMixin[Theme]):
    model = Theme

    async def get_default(self) -> Theme | None:
        statement = select(Theme).where(Theme.default == True)
        return await self.get_one_or_none(statement)



from fief.models.admin_api_key import AdminAPIKey
from fief.models.admin_session_token import AdminSessionToken
from fief.models.audit_log import AuditLog, AuditLogMessage
from fief.models.authorization_code import AuthorizationCode
from fief.models.base import MainBase, WorkspaceBase
from fief.models.client import Client, ClientType
from fief.models.email_template import EmailTemplate
from fief.models.generics import M_UUID, M
from fief.models.grant import Grant
from fief.models.login_session import LoginSession
from fief.models.oauth_account import OAuthAccount
from fief.models.oauth_provider import OAuthProvider
from fief.models.oauth_session import OAuthSession
from fief.models.permission import Permission
from fief.models.refresh_token import RefreshToken
from fief.models.registration_session import (
    RegistrationSession,
    RegistrationSessionFlow,
)
from fief.models.role import Role, RolePermission
from fief.models.session_token import SessionToken
from fief.models.tenant import Tenant
from fief.models.theme import Theme
from fief.models.user import User
from fief.models.user_field import UserField, UserFieldConfiguration, UserFieldType
from fief.models.user_field_value import UserFieldValue
from fief.models.user_permission import UserPermission
from fief.models.user_role import UserRole
from fief.models.webhook import Webhook
from fief.models.webhook_log import WebhookLog
from fief.models.workspace import Workspace
from fief.models.workspace_user import WorkspaceUser

__all__ = [
    "Workspace",
    "WorkspaceBase",
    "WorkspaceUser",
    "AdminAPIKey",
    "AdminSessionToken",
    "AuthorizationCode",
    "Client",
    "ClientType",
    "EmailTemplate",
    "Grant",
    "AuditLog",
    "AuditLogMessage",
    "LoginSession",
    "MainBase",
    "OAuthAccount",
    "OAuthProvider",
    "OAuthSession",
    "Permission",
    "RefreshToken",
    "RegistrationSession",
    "RegistrationSessionFlow",
    "Role",
    "RolePermission",
    "SessionToken",
    "Theme",
    "M",
    "M_UUID",
    "Tenant",
    "User",
    "UserField",
    "UserFieldConfiguration",
    "UserFieldType",
    "UserFieldValue",
    "UserPermission",
    "UserRole",
    "Webhook",
    "WebhookLog",
]



from typing import TYPE_CHECKING

from sqlalchemy import Boolean, Column, ForeignKey, String, Table
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fief.models.base import WorkspaceBase, get_prefixed_tablename
from fief.models.generics import CreatedUpdatedAt, UUIDModel
from fief.models.permission import Permission

if TYPE_CHECKING:
    from fief.models.user_permission import UserPermission

RolePermission = Table(
    get_prefixed_tablename("roles_permissions"),
    WorkspaceBase.metadata,
    Column(
        "role_id",
        ForeignKey(f"{get_prefixed_tablename('roles')}.id", ondelete="CASCADE"),
        primary_key=True,
    ),
    Column(
        "permission_id",
        ForeignKey(f"{get_prefixed_tablename('permissions')}.id", ondelete="CASCADE"),
        primary_key=True,
    ),
)


class Role(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "roles"

    name: Mapped[str] = mapped_column(String(length=255), nullable=False)
    granted_by_default: Mapped[bool] = mapped_column(
        Boolean, default=False, nullable=False
    )

    permissions: Mapped[list[Permission]] = relationship(
        "Permission", secondary=RolePermission, lazy="selectin"
    )
    user_permissions: Mapped[list["UserPermission"]] = relationship(
        "UserPermission", cascade="all, delete"
    )

    def __repr__(self) -> str:
        return f"Role(id={self.id}, name={self.name}, granted_by_default={self.granted_by_default}"



from collections.abc import Callable

from fastapi import Request, status
from fastapi.responses import JSONResponse

from fief.apps.auth.forms.auth import LoginForm
from fief.exceptions import (
    AuthorizeException,
    AuthorizeRedirectException,
    LoginException,
    LogoutException,
    OAuthException,
    TokenRequestException,
)
from fief.forms import FormHelper
from fief.services.authentication_flow import AuthenticationFlow
from fief.templates import templates

exception_handlers: dict[type[Exception], Callable] = {}


async def authorize_exception_handler(request: Request, exc: AuthorizeException):
    return templates.TemplateResponse(
        "auth/authorize.html",
        {
            "request": request,
            "error": exc.error.error_description,
            "tenant": exc.tenant,
            "fatal_error": True,
        },
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[AuthorizeException] = authorize_exception_handler


async def authorize_redirect_exception_handler(
    request: Request, exc: AuthorizeRedirectException
):
    return AuthenticationFlow.get_authorization_code_error_redirect(
        exc.redirect_uri,
        exc.response_mode,
        exc.error.error,
        error_description=exc.error.error_description,
        error_uri=exc.error.error_uri,
        state=exc.state,
    )


exception_handlers[AuthorizeRedirectException] = authorize_redirect_exception_handler


async def login_exception_handler(
    request: Request, exc: LoginException | OAuthException
):
    form_helper = FormHelper(
        LoginForm,
        "auth/login.html",
        request=request,
        context={"tenant": exc.tenant},
    )
    form = await form_helper.get_form()

    return templates.TemplateResponse(
        "auth/login.html",
        {
            "form": form,
            "request": request,
            "error": exc.error.error_description,
            "oauth_providers": exc.oauth_providers,
            "tenant": exc.tenant,
            "fatal_error": exc.fatal,
        },
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[LoginException] = login_exception_handler
exception_handlers[OAuthException] = login_exception_handler


async def token_request_exception_handler(request: Request, exc: TokenRequestException):
    return JSONResponse(
        status_code=status.HTTP_400_BAD_REQUEST,
        content=exc.error.dict(exclude_none=True),
    )


exception_handlers[TokenRequestException] = token_request_exception_handler


async def logout_exception_handler(request: Request, exc: LogoutException):
    return templates.TemplateResponse(
        "auth/logout.html",
        {
            "request": request,
            "error": exc.error.error_description,
            "tenant": exc.tenant,
            "fatal_error": True,
        },
        status_code=status.HTTP_400_BAD_REQUEST,
        headers={"X-Fief-Error": exc.error.error},
    )


exception_handlers[LogoutException] = logout_exception_handler

__all__ = ["exception_handlers"]



import functools

from fief.db.main import create_main_async_session_maker
from fief.db.workspace import WorkspaceEngineManager, get_workspace_session
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.users import get_user_db, get_user_manager
from fief.models import Client, User, Workspace, WorkspaceUser
from fief.repositories import (
    ClientRepository,
    TenantRepository,
    WorkspaceRepository,
    WorkspaceUserRepository,
)
from fief.schemas.user import UserCreateInternal
from fief.schemas.workspace import WorkspaceCreate
from fief.services.webhooks.trigger import trigger_webhooks
from fief.services.workspace_db import WorkspaceDatabase
from fief.settings import settings
from fief.tasks import send_task


class MainWorkspaceError(Exception):
    pass


class MainWorkspaceAlreadyExists(MainWorkspaceError):
    pass


class MainWorkspaceDoesNotExist(MainWorkspaceError):
    pass


class MainWorkspaceDoesNotHaveDefaultTenant(MainWorkspaceError):
    pass


class MainWorkspaceClientDoesNotExist(MainWorkspaceError):
    pass


class CreateMainFiefUserError(MainWorkspaceError):
    pass


async def get_main_fief_workspace() -> Workspace:
    main_async_session_maker = create_main_async_session_maker()
    async with main_async_session_maker() as session:
        workspace_repository = WorkspaceRepository(session)
        workspace = await workspace_repository.get_main()

        if workspace is None:
            raise MainWorkspaceDoesNotExist()

        return workspace


async def get_main_fief_client() -> Client:
    workspace = await get_main_fief_workspace()
    async with get_workspace_session(workspace, WorkspaceEngineManager()) as session:
        client_repository = ClientRepository(session)
        client = await client_repository.get_by_client_id(settings.fief_client_id)

        if client is None:
            raise MainWorkspaceClientDoesNotExist()

        return client


async def create_main_fief_workspace() -> Workspace:
    from fief.services.workspace_creation import WorkspaceCreation

    main_async_session_maker = create_main_async_session_maker()
    workspace_engine_manager = WorkspaceEngineManager()

    async with main_async_session_maker() as session:
        workspace_repository = WorkspaceRepository(session)
        workspace_user_repository = WorkspaceUserRepository(session)
        workspace = await workspace_repository.get_main()

        if workspace is not None:
            raise MainWorkspaceAlreadyExists()

        workspace_create = WorkspaceCreate(name="Fief")
        workspace_db = WorkspaceDatabase()
        workspace_creation = WorkspaceCreation(
            workspace_repository,
            workspace_user_repository,
            workspace_db,
            workspace_engine_manager,
        )

        workspace = await workspace_creation.create(
            workspace_create,
            default_domain=settings.fief_domain,
            default_client_id=settings.fief_client_id,
            default_client_secret=settings.fief_client_secret,
            default_encryption_key=settings.fief_encryption_key,
            default_redirect_uris=[
                f"http://{settings.fief_domain}/admin/auth/callback",
                f"https://{settings.fief_domain}/admin/auth/callback",
                "http://localhost:8000/docs/oauth2-redirect",
            ],
        )

    return workspace


async def create_main_fief_user(email: str, password: str | None = None) -> User:
    workspace = await get_main_fief_workspace()
    main_async_session_maker = create_main_async_session_maker()
    async with main_async_session_maker() as session:
        workspace_user_repository = WorkspaceUserRepository(session)

        async with get_workspace_session(
            workspace, WorkspaceEngineManager()
        ) as session:
            tenant_repository = TenantRepository(session)
            tenant = await tenant_repository.get_default()

            if tenant is None:
                raise MainWorkspaceDoesNotHaveDefaultTenant()

            user_db = await get_user_db(session, tenant)
            audit_logger = await get_audit_logger(workspace, None, None)
            user_manager = await get_user_manager(
                user_db,
                tenant,
                workspace,
                send_task,
                audit_logger,
                functools.partial(
                    trigger_webhooks, workspace_id=workspace.id, send_task=send_task
                ),
            )
            if password is None:
                password = user_manager.password_helper.generate()
            user = await user_manager.create_with_fields(
                UserCreateInternal(email=email, password=password, tenant_id=tenant.id),
                user_fields=[],
            )

        workspace_user = WorkspaceUser(workspace_id=workspace.id, user_id=user.id)
        await workspace_user_repository.create(workspace_user)

    return user



import functools

import sentry_sdk
from fastapi import FastAPI, status
from fastapi.responses import RedirectResponse
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from sentry_sdk.integrations.redis import RedisIntegration

from fief import __version__
from fief.apps import api_app, auth_app, dashboard_app
from fief.db.workspace import get_workspace_session
from fief.dependencies.db import main_async_session_maker, workspace_engine_manager
from fief.logger import init_audit_logger, logger
from fief.settings import settings

sentry_sdk.init(
    dsn=settings.sentry_dsn_server,
    environment=settings.environment.value,
    traces_sample_rate=0.1,
    release=__version__,
    integrations=[RedisIntegration()],
)

app = FastAPI(openapi_url=None)

app.add_middleware(SentryAsgiMiddleware)


@app.get("/admin")
async def get_admin():
    return RedirectResponse("/admin/", status_code=status.HTTP_308_PERMANENT_REDIRECT)


app.mount("/admin/api", api_app)
app.mount("/admin", dashboard_app)
app.mount("/", auth_app)


@app.on_event("startup")
async def on_startup():
    init_audit_logger(
        main_async_session_maker,
        functools.partial(
            get_workspace_session, workspace_engine_manager=workspace_engine_manager
        ),
    )
    logger.info("Fief Server started", version=__version__)


@app.on_event("shutdown")
async def on_shutdown():
    await workspace_engine_manager.close_all()


__all__ = ["app"]



from typing import Protocol


class EmailProvider(Protocol):
    def send_email(
        self,
        *,
        sender: tuple[str, str | None],
        recipient: tuple[str, str | None],
        subject: str,
        html: str | None = None,
        text: str | None = None,
    ):
        ...


class EmailError(Exception):
    def __init__(self, message: str):
        self.message = message


class SendEmailError(EmailError):
    pass


def format_address(email: str, name: str | None = None):
    return email if name is None else f"{name} <{email}>"



"""Add Role, Permission, UserRole and UserPermission models

Revision ID: 7f4cbde3bcb7
Revises: ac76d4c912eb
Create Date: 2022-05-31 13:21:15.383928

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "7f4cbde3bcb7"
down_revision = "ac76d4c912eb"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_permissions",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("codename", sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("codename"),
    )
    op.create_index(
        op.f("ix_fief_permissions_created_at"),
        "fief_permissions",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_permissions_updated_at"),
        "fief_permissions",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_roles",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("granted_by_default", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_roles_created_at"), "fief_roles", ["created_at"], unique=False
    )
    op.create_index(
        op.f("ix_fief_roles_updated_at"), "fief_roles", ["updated_at"], unique=False
    )
    op.create_table(
        "roles_permissions",
        sa.Column("role_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("permission_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["permission_id"], ["fief_permissions.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["role_id"], ["fief_roles.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("role_id", "permission_id"),
    )
    op.create_table(
        "fief_user_permissions",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("permission_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("from_role_id", fief.models.generics.GUID(), nullable=True),
        sa.ForeignKeyConstraint(
            ["from_role_id"], ["fief_roles.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["permission_id"], ["fief_permissions.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["user_id"], ["fief_users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "permission_id", "from_role_id"),
    )
    op.create_index(
        op.f("ix_fief_user_permissions_created_at"),
        "fief_user_permissions",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_user_permissions_updated_at"),
        "fief_user_permissions",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_user_roles",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("role_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(["role_id"], ["fief_roles.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_id"], ["fief_users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "role_id"),
    )
    op.create_index(
        op.f("ix_fief_user_roles_created_at"),
        "fief_user_roles",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_user_roles_updated_at"),
        "fief_user_roles",
        ["updated_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_fief_user_roles_updated_at"), table_name="fief_user_roles")
    op.drop_index(op.f("ix_fief_user_roles_created_at"), table_name="fief_user_roles")
    op.drop_table("fief_user_roles")
    op.drop_index(
        op.f("ix_fief_user_permissions_updated_at"), table_name="fief_user_permissions"
    )
    op.drop_index(
        op.f("ix_fief_user_permissions_created_at"), table_name="fief_user_permissions"
    )
    op.drop_table("fief_user_permissions")
    op.drop_table("roles_permissions")
    op.drop_index(op.f("ix_fief_roles_updated_at"), table_name="fief_roles")
    op.drop_index(op.f("ix_fief_roles_created_at"), table_name="fief_roles")
    op.drop_table("fief_roles")
    op.drop_index(op.f("ix_fief_permissions_updated_at"), table_name="fief_permissions")
    op.drop_index(op.f("ix_fief_permissions_created_at"), table_name="fief_permissions")
    op.drop_table("fief_permissions")
    # ### end Alembic commands ###



from collections.abc import Generator
from contextlib import contextmanager

from alembic import command
from alembic.config import Config
from alembic.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from sqlalchemy import create_engine, exc, inspect
from sqlalchemy.engine import Engine
from sqlalchemy.schema import CreateSchema

from fief.db.types import DatabaseConnectionParameters
from fief.paths import ALEMBIC_CONFIG_FILE


class WorkspaceDatabaseError(Exception):
    pass


class WorkspaceDatabaseConnectionError(WorkspaceDatabaseError):
    def __init__(self, message: str) -> None:
        self.message = message


class WorkspaceDatabase:
    def migrate(
        self,
        database_connection_parameters: DatabaseConnectionParameters,
        schema_name: str,
    ) -> str:
        try:
            self._ensure_schema(database_connection_parameters, schema_name)
            with self._get_engine(
                database_connection_parameters, schema_name
            ) as engine:
                with engine.begin() as connection:
                    config = self._get_alembic_base_config()
                    config.attributes["configure_logger"] = False
                    config.attributes["connection"] = connection
                    command.upgrade(config, "head")
        except exc.OperationalError as e:
            raise WorkspaceDatabaseConnectionError(str(e)) from e
        return self.get_latest_revision()

    def check_connection(
        self, database_connection_parameters: DatabaseConnectionParameters
    ) -> tuple[bool, str | None]:
        with self._get_engine(database_connection_parameters) as engine:
            try:
                with engine.begin():
                    return True, None
            except exc.OperationalError as e:
                return False, str(e)

    def get_latest_revision(self) -> str:
        config = self._get_alembic_base_config()
        script = ScriptDirectory.from_config(config)
        with EnvironmentContext(config, script) as environment:
            return str(environment.get_head_revision())

    def _get_alembic_base_config(self) -> Config:
        return Config(ALEMBIC_CONFIG_FILE, ini_section="workspace")

    @contextmanager
    def _get_engine(
        self,
        database_connection_parameters: DatabaseConnectionParameters,
        schema_name: str | None = None,
    ) -> Generator[Engine, None, None]:
        database_url, connect_args = database_connection_parameters

        if schema_name is not None:
            dialect_name = database_url.get_dialect().name
            if dialect_name == "postgresql":
                connect_args["options"] = f"-csearch_path={schema_name}"
            elif dialect_name == "mysql":
                database_url = database_url.set(database=schema_name)

        engine = create_engine(database_url, connect_args=connect_args)

        yield engine

        engine.dispose()

    def _ensure_schema(
        self,
        database_connection_parameters: DatabaseConnectionParameters,
        schema_name: str,
    ):
        with self._get_engine(database_connection_parameters) as engine:
            dialect_name = engine.dialect.name
            if dialect_name == "sqlite":
                return

            inspector = inspect(engine)
            schemas = inspector.get_schema_names()
            if schema_name not in schemas:
                with engine.begin() as connection:
                    connection.execute(CreateSchema(schema_name))



import secrets
from typing import Literal

from jwcrypto import jwk

from fief.settings import settings


class AuthException(Exception):
    pass


class InvalidAccessToken(AuthException):
    pass


def generate_jwk(kid: str, use: Literal["sig", "enc"]) -> jwk.JWK:
    return jwk.JWK.generate(
        kty="RSA", size=settings.generated_jwk_size, use=use, kid=kid
    )


def load_jwk(json: str) -> jwk.JWK:
    return jwk.JWK.from_json(json)


def generate_signature_jwk_string() -> str:
    key = generate_jwk(secrets.token_urlsafe(), "sig")
    return key.export()



"""Reduce size of OAuthAccount.account_id and OAuthAccount.account_email for MySQL compat

Revision ID: ab9cf21c76dd
Revises: aa662768d69b
Create Date: 2022-11-29 10:10:43.160013

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "ab9cf21c76dd"
down_revision = "aa662768d69b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    if connection.dialect.name == "sqlite":
        with op.batch_alter_table("fief_oauth_accounts") as batch_op:
            batch_op.alter_column(
                "account_id",
                type_=sa.String(512),
                existing_type=sa.String(1024),
                existing_nullable=False,
            )
            batch_op.alter_column(
                "account_email",
                type_=sa.String(512),
                existing_type=sa.String(1024),
                existing_nullable=False,
            )
    else:
        op.alter_column(
            "fief_oauth_accounts",
            "account_id",
            type_=sa.String(512),
            existing_type=sa.String(1024),
            existing_nullable=False,
        )
        op.alter_column(
            "fief_oauth_accounts",
            "account_email",
            type_=sa.String(512),
            existing_type=sa.String(1024),
            existing_nullable=True,
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
